# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.tests.loader import TestLoader, LoaderException, _requests_session

import tests.ducktape_mock

import os
import os.path
import pytest
import re
import requests
import tempfile
import yaml

from mock import Mock
from requests_testadapter import Resp


class LocalFileAdapter(requests.adapters.HTTPAdapter):
    def build_response_from_file(self, request):
        file_path = request.url[7:]
        with open(file_path, 'rb') as file:
            buff = bytearray(os.path.getsize(file_path))
            file.readinto(buff)
            resp = Resp(buff)
            r = self.build_response(request, resp)

            return r

    def send(self, request, stream=False, timeout=None,
             verify=True, cert=None, proxies=None):

        return self.build_response_from_file(request)


def resources_dir():
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources")


def discover_dir():
    """Return the absolute path to the directory to use with discovery tests."""
    return os.path.join(resources_dir(), "loader_test_directory")


def sub_dir_a():
    return os.path.join(discover_dir(), "sub_dir_a")


def num_tests_in_file(fpath):
    """Count expected number of tests in the file.
    Search for NUM_TESTS = N

    return N if pattern is present else 0
    """
    with open(fpath, 'r') as fd:
        match = re.search(r'^NUM_TESTS\s*=\s*(\d+)', fd.read(), re.MULTILINE)

        if not match:
            return 0
        return int(match.group(1))


def num_tests_in_dir(dpath):
    """Walk through directory subtree and count up expected number of tests that TestLoader should find."""
    assert os.path.exists(dpath)
    assert os.path.isdir(dpath)

    num_tests = 0
    for pwd, dirs, files in os.walk(dpath):
        for f in files:
            if not f.endswith('.py'):
                continue
            file_path = os.path.abspath(os.path.join(pwd, f))
            num_tests += num_tests_in_file(file_path)
    return num_tests


def invalid_test_suites():
    dpath = os.path.join(discover_dir(), 'invalid_test_suites')
    params = []
    for pwd, dirs, files in os.walk(dpath):
        for f in files:
            if not f.endswith('.yml'):
                continue
            file_path = os.path.abspath(os.path.join(pwd, f))
            params.append(pytest.param(file_path, id=os.path.basename(file_path)))
    return params


class CheckTestLoader(object):
    def setup_method(self, method):
        self.SESSION_CONTEXT = tests.ducktape_mock.session_context()
        # To simplify unit tests, add file:// support to the test loader's functionality for loading previous
        # report.json files
        _requests_session.mount('file://', LocalFileAdapter())

    @pytest.mark.parametrize('suite_file_path', invalid_test_suites())
    def check_test_loader_raises_on_invalid_test_suite(self, suite_file_path):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        with pytest.raises(LoaderException):
            loader.load([suite_file_path])

    @pytest.mark.parametrize(['expected_count', 'input_symbols', 'excluded_symbols'], [
        pytest.param(6, [
            # import a single test suite with an import statement
            os.path.join(discover_dir(), 'test_suite_with_single_import.yml')
        ], None, id='load dependency path'),
        pytest.param(2, [
            # test that a self import doesn't cause an infinite loop
            os.path.join(discover_dir(), 'test_suite_with_self_import.yml')
        ], None, id='self load in import'),
        pytest.param(5, [
            # test that a cyclic import doesn't cause an infinite loop
            os.path.join(discover_dir(), 'test_suite_cyclic_b.yml')
        ], None, id='self load in import'),
        pytest.param(5, [
            # test that a cyclic import starting with second import
            os.path.join(discover_dir(), 'test_suite_cyclic_a.yml')
        ], None, id='self load in import'),
        pytest.param(5, [
            # test single import statement with parent reference
            os.path.join(discover_dir(), 'test_suites', 'sub_dir_test_import.yml')
        ], None, id='self load in import'),
        pytest.param(8, [
            # see test suite files for number of tests in it.
            # decorated test suite includes 2 tests;
            # single includes 4 tests; multiple includes 3 tests;
            # however both single and multiple test suites include one test method of test_by.py,
            # so total -= 1
            os.path.join(discover_dir(), 'test_suite_single.yml'),
            os.path.join(discover_dir(), 'test_suite_multiple.yml'),
            os.path.join(discover_dir(), 'test_suite_decorated.yml'),
        ], None, id='load multiple test suite files'),
        pytest.param(5, [
            # see test suite file for number of tests in it
            os.path.join(discover_dir(), 'test_suites', 'test_suite_glob.yml')
        ], None, id='load test suite with globs'),
        pytest.param(2, [
            # test suite that includes sub_dir_a/test_c.py (1 test total):
            os.path.join(discover_dir(), 'test_suites', 'sub_dir_a_test_c.yml'),
            # explicitly include test_a.yml (1 test total)
            os.path.join(discover_dir(), 'test_a.py')
        ], None, id='load both file and suite'),
        pytest.param(1, [
            # test suite that includes sub_dir_a/test_c.py (1 test total):
            os.path.join(discover_dir(), 'test_suites', 'sub_dir_a_test_c.yml'),
            # explicitly include test_a.yml (1 test total)
            os.path.join(discover_dir(), 'test_a.py')
        ], [
            # explicitly exclude the sub_dir_a/test_c.py (included with test suite):
            os.path.join(sub_dir_a(), 'test_c.py'),
        ], id='global exclude overrides test suite include'),
        pytest.param(4, [
            # sub_dir_a contains 4 total tests
            # test suite that includes sub_dir_a/*.py but excludes sub_dir_a/test_d.py:
            os.path.join(discover_dir(), 'test_suites', 'sub_dir_a_with_exclude.yml'),
            # explicitly include sub_dir_a/test_d.py to override exclusion from test suite:
            os.path.join(sub_dir_a(), 'test_d.py')
        ], None, id='global include overrides test suite exclude'),
        pytest.param(1, [
            # load two test suites and two files that all point to the same actual test
            # and verify that in the end only 1 test has been loaded
            os.path.join(discover_dir(), 'test_suites', 'sub_dir_a_test_c.yml'),
            os.path.join(discover_dir(), 'test_suites', 'sub_dir_a_test_c_via_class.yml'),
            os.path.join(sub_dir_a(), 'test_c.py'),
            os.path.join(sub_dir_a(), 'test_c.py::TestC')
        ], None, id='same test in test suites and test files')
    ])
    def check_test_loader_with_test_suites_and_files(self, expected_count, input_symbols, excluded_symbols):
        """
        When both files and test suites are loaded, files (both included and excluded) are
        loaded after and separately from the test suites, so even if a test suite excludes file A,
        it will be included if it's passed directly. And if file A is excluded directly, even if any of
        the test suites includes it, it will still be excluded.
        """
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        tests = loader.load(input_symbols, excluded_test_symbols=excluded_symbols)
        assert len(tests) == expected_count

    def check_test_loader_with_directory(self):
        """Check discovery on a directory."""
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        tests = loader.load([discover_dir()])
        assert len(tests) == num_tests_in_dir(discover_dir())

    @pytest.mark.parametrize(['dir_', 'file_name'], [
        pytest.param(discover_dir(), 'test_a.py'),
        pytest.param(resources_dir(), 'a.py')
    ])
    def check_test_loader_with_file(self, dir_, file_name):
        """Check discovery on a file. """
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        module_path = os.path.join(dir_, file_name)

        tests = loader.load([module_path])
        assert len(tests) == num_tests_in_file(module_path)

    def check_test_loader_with_glob(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        file_glob = os.path.join(discover_dir(), "*_a.py")  # should resolve to test_a.py only
        tests = loader.load([file_glob])
        assert len(tests) == 1

    def check_test_loader_multiple_files(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        file_a = os.path.join(discover_dir(), "test_a.py")
        file_b = os.path.join(discover_dir(), "test_b.py")

        tests = loader.load([file_a, file_b])
        assert len(tests) == num_tests_in_file(file_a) + num_tests_in_file(file_b)

    def check_test_loader_include_dir_exclude_file(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        excluded_file_a = os.path.join(discover_dir(), "test_a.py")
        excluded_file_b = os.path.join(discover_dir(), "test_b.py")
        num_excluded = num_tests_in_file(excluded_file_a) + num_tests_in_file(excluded_file_b)
        tests = loader.load([discover_dir()], [excluded_file_a, excluded_file_b])
        assert len(tests) == num_tests_in_dir(discover_dir()) - num_excluded

    def check_test_loader_exclude_subdir(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        included_dir = discover_dir()
        excluded_dir = sub_dir_a()
        tests = loader.load([included_dir], [excluded_dir])
        assert len(tests) == num_tests_in_dir(included_dir) - num_tests_in_dir(excluded_dir)

    def check_test_loader_exclude_subdir_glob(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        included_dir = discover_dir()
        excluded_dir = sub_dir_a()
        excluded_dir_glob = os.path.join(sub_dir_a(), "*.py")
        tests = loader.load([included_dir], [excluded_dir_glob])
        assert len(tests) == num_tests_in_dir(included_dir) - num_tests_in_dir(excluded_dir)

    def check_test_loader_raises_when_nothing_is_included(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        file_a = os.path.join(discover_dir(), "test_a.py")
        file_b = os.path.join(discover_dir(), "test_b.py")
        with pytest.raises(LoaderException):
            loader.load([file_a, file_b], [discover_dir()])

    def check_test_loader_raises_on_include_subdir_exclude_parent_dir(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        with pytest.raises(LoaderException):
            loader.load([(sub_dir_a())], [(discover_dir())])

    def check_test_loader_with_nonexistent_file(self):
        """Check discovery on a non-existent path should throw LoaderException"""
        with pytest.raises(LoaderException):
            loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
            loader.load([os.path.join(discover_dir(), "file_that_does_not_exist.py")])

    def check_test_loader_include_dir_without_tests(self):
        with pytest.raises(LoaderException):
            loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
            loader.load([os.path.join(discover_dir(), "sub_dir_no_tests")])

    def check_test_loader_include_file_without_tests(self):
        with pytest.raises(LoaderException):
            loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
            loader.load([os.path.join(discover_dir(), "sub_dir_no_tests", "just_some_file.py")])

    def check_test_loader_allow_exclude_dir_without_tests(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        tests = loader.load([discover_dir()], [os.path.join(discover_dir(), "sub_dir_no_tests")])
        assert len(tests) == num_tests_in_dir(discover_dir())

    def check_test_loader_allow_exclude_file_without_tests(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        tests = loader.load([discover_dir()], [os.path.join(discover_dir(), "sub_dir_no_tests", "just_some_file.py")])
        assert len(tests) == num_tests_in_dir(discover_dir())

    def check_test_loader_allow_exclude_nonexistent_file(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        tests = loader.load([discover_dir()], [os.path.join(discover_dir(), "file_that_does_not_exist.py")])
        assert len(tests) == num_tests_in_dir(discover_dir())

    def check_test_loader_with_class(self):
        """Check test discovery with discover class syntax."""
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        tests = loader.load([os.path.join(discover_dir(), "test_b.py::TestBB")])
        assert len(tests) == 2

        # Sanity check, test that it discovers two test class & 3 tests if it searches the whole module
        tests = loader.load([os.path.join(discover_dir(), "test_b.py")])
        assert len(tests) == 3

    def check_test_loader_include_dir_exclude_class(self):
        """Check test discovery with discover class syntax."""
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        tests = loader.load([discover_dir()], [os.path.join(discover_dir(), "test_b.py::TestBB")])
        # TestBB contains 2 test methods
        assert len(tests) == num_tests_in_dir(discover_dir()) - 2

    def check_test_loader_include_class_exclude_method(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        included = [os.path.join(discover_dir(), "test_b.py::TestBB")]
        excluded = [os.path.join(discover_dir(), "test_b.py::TestBB.test_bb_one")]
        tests = loader.load(included, excluded)
        # TestBB contains 2 test methods, but 1 is excluded
        assert len(tests) == 1

    def check_test_loader_include_dir_exclude_method(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        excluded = [os.path.join(discover_dir(), "test_b.py::TestBB.test_bb_one")]
        tests = loader.load([discover_dir()], excluded)
        # excluded 1 method only
        assert len(tests) == num_tests_in_dir(discover_dir()) - 1

    def check_test_loader_with_matrix_params(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        included = [os.path.join(discover_dir(), 'test_decorated.py::TestMatrix.test_thing@{"x": 1,"y": "test "}')]
        tests = loader.load(included)
        # TestMatrix contains a single parametrized method. Since we only provide a single set of params, we should
        # end up with a single context
        assert len(tests) == 1
        assert tests[0].injected_args == {'x': 1, 'y': 'test '}

    def check_test_loader_with_params_special_chars(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        included = [os.path.join(discover_dir(
        ), r'test_decorated.py::TestParametrizdeSpecial.test_special_characters_params'
           r'@{"version": "6.1.0", "chars": "!@#$%^&*()_+::.,/? \"{}\\"}')]
        tests = loader.load(included)
        # TestMatrix contains a single parametrized method. Since we only provide a single set of params, we should
        # end up with a single context
        assert len(tests) == 1
        assert tests[0].injected_args == {'version': '6.1.0', 'chars': '!@#$%^&*()_+::.,/? \"{}\\'}

    def check_test_loader_with_multiple_matrix_params(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        params = '[{"x": 1,"y": "test "}, {"x": 2,"y": "I\'m"}]'
        included = [os.path.join(discover_dir(), 'test_decorated.py::TestMatrix.test_thing@{}'.format(params))]
        tests = loader.load(included)
        # TestMatrix contains a single parametrized method.
        # We provide two sets of params, so we should end up with two contexts
        assert len(tests) == 2
        injected_args = map(lambda t: t.injected_args, tests)
        assert {'x': 1, 'y': 'test '} in injected_args
        assert {'x': 2, 'y': "I'm"} in injected_args

    def check_test_loader_with_parametrize(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        included = [os.path.join(discover_dir(), 'test_decorated.py::TestParametrized.test_thing@{"x":1,"y":2}')]
        tests = loader.load(included)
        assert len(tests) == 1
        assert tests[0].injected_args == {'x': 1, 'y': 2}

    def check_test_loader_with_parametrize_with_objects(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        parameters = '{"d": {"a": "A"}, "lst": ["whatever"]}'
        included = [os.path.join(
            discover_dir(), 'test_decorated.py::TestObjectParameters.test_thing@{}'.format(parameters))]
        tests = loader.load(included)
        assert len(tests) == 1
        assert tests[0].injected_args == {'d': {'a': 'A'}, 'lst': ['whatever']}

    def check_test_loader_with_injected_args(self):
        """When the --parameters command-line option is used, the loader behaves a little bit differently:

        each test method annotated with @parametrize or @matrix should only expand to a single discovered test,
        and the injected args should be those passed in from command-line.
        """
        # parameter values don't have to match any of the pre-defined parameters in the annotation
        # moreover, even parameter keys don't have to match method arguments, though if that's the case
        # the runner will complain, but the loader wouldn't care (this has been ducktape's behavior for a while now)
        injected_args = {"x": 100, "y": -100}
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock(), injected_args=injected_args)

        file = os.path.join(discover_dir(), "test_decorated.py")
        tests = loader.load([file])
        assert len(tests) == 6

        for t in tests:
            assert t.injected_args == injected_args

    def check_test_loader_raises_with_both_injected_args_and_parameters(self):
        """One should not use both --parameters command-line flag and parameterized test symbols at the same time.
        Loader will explicitly raise in such cases to avoid confusing behavior.
        """
        injected_args = {"x": 1, "y": "test "}
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock(), injected_args=injected_args)
        included = [os.path.join(discover_dir(), 'test_decorated.py::TestMatrix.test_thing@{"x": 1,"y": "test "}')]
        with pytest.raises(LoaderException, match='Cannot use both'):
            loader.load(included)

    def check_test_loader_raises_on_params_not_found(self):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        # parameter syntax is valid, but there is no such parameter defined in the test annotation in the code
        included = [os.path.join(discover_dir(), 'test_decorated.py::TestMatrix.test_thing@{"x": 1,"y": "missing"}')]
        with pytest.raises(LoaderException, match='No tests to run'):
            loader.load(included)

    @pytest.mark.parametrize("symbol", [
        # no class
        'test_decorated.py::.test_thing'
        # no method
        'test_decorated.py::TestMatrix@{"x": 1, "y": "test "}'
        # invalid json in params
        'test_decorated.py::TestMatrix.test_thing@{x: 1,"y": "test "}'
    ])
    def check_test_loader_raises_on_malformed_test_discovery_symbol(self, symbol):
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        included = [os.path.join(discover_dir(), symbol)]
        with pytest.raises(LoaderException, match='Invalid discovery symbol'):
            loader.load(included)

    def check_test_loader_exclude_with_injected_args(self):
        injected_args = {"x": 1, "y": -1}
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock(), injected_args=injected_args)

        included = [os.path.join(discover_dir(), "test_decorated.py")]
        excluded = [os.path.join(discover_dir(), "test_decorated.py::TestStackedMatrix")]
        tests = loader.load(included, excluded)
        # test_decorated.py contains 5 test methods total
        # we exclude 1 class with 1 method so should be 4
        # exclusion shouldn't care about injected args
        assert len(tests) == 5

        for t in tests:
            assert t.injected_args == injected_args

    def check_test_loader_exclude_with_params(self):
        """
        Checks behavior of exclude flag with parametrized annotations.
        Should exclude only a single permutation of the method
        """
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        # included 8 tests
        included = [os.path.join(discover_dir(), "test_decorated.py::TestMatrix")]
        # exclude 1 test
        excluded = [os.path.join(discover_dir(), 'test_decorated.py::TestMatrix.test_thing@{"x": 1,"y": "test "}')]
        tests = loader.load(included, excluded)
        assert len(tests) == 7

    def check_test_loader_exclude_with_params_multiple(self):
        """
        Checks behavior of exclude flag with parametrized annotations.
        Should exclude two permutations of the method
        """
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        # include 8 tests
        included = [os.path.join(discover_dir(), "test_decorated.py::TestMatrix")]
        # exclude 2 tests
        params = '[{"x": 1,"y": "test "}, {"x": 2,"y": "I\'m"}]'
        excluded = [os.path.join(discover_dir(), 'test_decorated.py::TestMatrix.test_thing@{}'.format(params))]
        tests = loader.load(included, excluded)
        assert len(tests) == 6

    def check_test_loader_with_subsets(self):
        """Check that computation of subsets work properly. This validates both that the division of tests is correct
        (i.e. as even a distribution as we can get but uneven in the expected way when necessary) and that the division
        happens after the expansion of tests marked for possible expansion (e.g. matrix, parametrize)."""

        file = os.path.join(discover_dir(), "test_decorated.py")

        # The test file contains 18 tests. With 4 subsets, first two subset should have an "extra"
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock(), subset=0, subsets=4)
        tests = loader.load([file])
        assert len(tests) == 5

        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock(), subset=1, subsets=4)
        tests = loader.load([file])
        assert len(tests) == 5

        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock(), subset=2, subsets=4)
        tests = loader.load([file])
        assert len(tests) == 4

        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock(), subset=3, subsets=4)
        tests = loader.load([file])
        assert len(tests) == 4

    def check_test_loader_with_invalid_subsets(self):
        """Check that the TestLoader throws an exception if the requests subset is larger than the number of subsets"""
        with pytest.raises(ValueError):
            TestLoader(self.SESSION_CONTEXT, logger=Mock(), subset=4, subsets=4)
        with pytest.raises(ValueError):
            TestLoader(self.SESSION_CONTEXT, logger=Mock(), subset=5, subsets=4)

    def check_test_loader_with_time_based_subsets(self):
        """Check that computation of subsets using a report with timing information correctly generates subsets that
        are optimized based on timing rather than number of tests.
        """

        file = os.path.join(discover_dir(), "test_b.py")
        report_url = "file://" + os.path.join(resources_dir(), "report.json")

        # The expected behavior of the current implementation is to add tests to each subset from largest to smallest,
        # using the least full subset each time. The test data with times of (10, 5, 1) should result in the first
        # subset containing 1 test and the second containing 2 (the opposite of the simple count-based strategy)

        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock(), subset=0, subsets=2, historical_report=report_url)
        tests = loader.load([file])
        assert len(tests) == 1

        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock(), subset=1, subsets=2, historical_report=report_url)
        tests = loader.load([file])
        assert len(tests) == 2

    def check_loader_with_non_yml_file(self):
        """
        test loading a test file as an import
        """
        file = os.path.join(discover_dir(), 'test_suite_import_py.yml')
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        with pytest.raises(LoaderException, match=r'Failed to load test suite from file: \S+test_a\.py'):
            loader.load([file])

    def check_loader_with_non_suite_yml_file(self):
        """
        test importing a suite that is malformed
        """
        file1 = os.path.join(discover_dir(), 'test_suite_malformed.yml')
        file2 = os.path.join(discover_dir(), 'test_suite_import_malformed.yml')
        loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
        with pytest.raises(LoaderException, match='No tests found in  simple_malformed_suite'):
            loader.load([file1])
        with pytest.raises(LoaderException, match='No tests found in  simple_malformed_suite'):
            loader.load([file2])

    def check_test_loader_with_absolute_path(self):
        """
        Test loading suites using absolute paths to other imported suites as well as absolute paths
        to tests in the suite
        """
        with tempfile.TemporaryDirectory() as td:
            temp_suite1 = os.path.join(td, 'temp_suite1.yml')
            temp_suite2 = os.path.join(td, 'temp_suite2.yml')
            with open(temp_suite1, 'w') as f:
                test_yaml1 = yaml.dump({'import': str(os.path.join(td, 'temp_suite2.yml')),
                                        'suite': [os.path.abspath(os.path.join(discover_dir(), "test_a.py"))]})
                f.write(test_yaml1)

            with open(temp_suite2, 'w') as f:
                test_yaml2 = yaml.dump({'suite': [os.path.abspath(os.path.join(discover_dir(), "test_b.py"))]})
                f.write(test_yaml2)

            loader = TestLoader(self.SESSION_CONTEXT, logger=Mock())
            tests = loader.load([temp_suite1])
            assert len(tests) == 4


def join_parsed_symbol_components(parsed):
    """
    Join together a parsed symbol

    e.g.
        {
            'path': 'path/to/dir/test_file.py',
            'cls': 'ClassName',
            'method': 'method'
        },
        ->
        'path/to/dir/test_file.py::ClassName.method'
    """
    symbol = os.path.join(parsed['path'])

    if parsed['cls'] or parsed['method']:
        symbol += "::"
        symbol += parsed['cls']
        if parsed['method']:
            symbol += "."
            symbol += parsed['method']

    return symbol


def normalize_ending_slash(dirname):
    if dirname.endswith(os.path.sep):
        dirname = dirname[:-len(os.path.sep)]
    return dirname


class CheckParseSymbol(object):
    def check_parse_discovery_symbol(self):
        """Check that "test discovery symbol" parsing logic works correctly"""
        parsed_symbols = [
            {
                'path': 'path/to/dir',
                'cls': '',
                'method': ''
            },
            {
                'path': 'path/to/dir/test_file.py',
                'cls': '',
                'method': ''
            },
            {
                'path': 'path/to/dir/test_file.py',
                'cls': 'ClassName',
                'method': ''
            },
            {
                'path': 'path/to/dir/test_file.py',
                'cls': 'ClassName',
                'method': 'method'
            },
            {
                'path': 'path/to/dir',
                'cls': 'ClassName',
                'method': ''
            },
            {
                'path': 'test_file.py',
                'cls': '',
                'method': ''
            },
            {
                'path': 'test_file.py',
                'cls': 'ClassName',
                'method': ''
            },
            {
                'path': 'test_file.py',
                'cls': 'ClassName',
                'method': 'method'
            }
        ]

        loader = TestLoader(tests.ducktape_mock.session_context(), logger=Mock())
        for parsed in parsed_symbols:
            symbol = join_parsed_symbol_components(parsed)

            expected_parsed = (
                normalize_ending_slash(parsed['path']),
                parsed['cls'],
                parsed['method']
            )

            actually_parsed = loader._parse_discovery_symbol(symbol)
            actually_parsed = (
                normalize_ending_slash(actually_parsed[0]),
                actually_parsed[1],
                actually_parsed[2]
            )

            assert actually_parsed == expected_parsed, "%s did not parse as expected" % symbol



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.command_line.parse_args import parse_args

from six.moves import cStringIO as StringIO

import os
import re
import shutil
import sys
import tempfile


class Capturing(object):
    """This context manager can be used to capture stdout from a function call.
    E.g.
        with Capture() as captured:
            call_function()
        assert captured.output == expected_output
    """

    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.output = self._stringio.getvalue()
        sys.stdout = self._stdout


class CheckParseArgs(object):

    def check_empty_args(self):
        """Check that parsing an empty args list results in printing a usage message, followed by sys.exit(0) """
        try:
            with Capturing() as captured:
                parse_args([])
        except SystemExit as e:
            assert e.code == 0
            assert captured.output.find("usage") >= 0

    def check_version(self):
        """If --version is present, ducktape should print version and exit"""
        try:
            with Capturing() as captured:
                parse_args(["--version"])
        except SystemExit as e:
            assert e.code == 0
            assert re.search(r"[\d]+\.[\d]+\.[\d]+", captured.output) is not None

    def check_empty_test_path(self):
        """Check that default test_path is an array consisting of cwd."""
        args = ["--collect-only"]
        parsed = parse_args(args)
        parsed_paths = [os.path.abspath(p) for p in parsed["test_path"]]

        assert parsed_paths == [os.path.abspath('.')]

    def check_multiple_test_paths(self):
        """Check that parser properly handles multiple "test paths". It should capture a list of test paths. """
        paths = ["path1"]
        args = ["--debug"] + paths + ["--collect-only"]
        parsed = parse_args(args)
        assert parsed["test_path"] == paths

        paths = ["path%d" % i for i in range(10)]
        args = ["--cluster-file", "my-cluster-file"] + paths + ["--debug", "--exit-first"]
        parsed = parse_args(args)
        assert parsed["test_path"] == paths

    def check_multiple_exclude(self):
        excluded = ["excluded1", "excluded2"]
        args = ["--collect-only", "--exclude"] + excluded + ["--debug"]
        parsed = parse_args(args)
        assert parsed["exclude"] == excluded

    def check_config_overrides(self, monkeypatch):
        """Check that parsed arguments pick up values from config files, and that overrides match precedence."""

        tmpdir = tempfile.mkdtemp(dir="/tmp")
        # Create tmp file for global config
        project_cfg_filename = os.path.join(tmpdir, "ducktape-project.cfg")
        user_cfg_filename = os.path.join(tmpdir, "ducktape-user.cfg")

        project_cfg = [
            "--cluster-file CLUSTERFILE-project",
            "--results-root RESULTSROOT-project",
            "--parameters PARAMETERS-project"
        ]

        # user_cfg options should override project_cfg
        user_cfg = [
            "--results-root RESULTSROOT-user",
            "--parameters PARAMETERS-user"
        ]

        try:
            monkeypatch.setattr("ducktape.command_line.defaults.ConsoleDefaults.PROJECT_CONFIG_FILE",
                                project_cfg_filename)
            monkeypatch.setattr("ducktape.command_line.defaults.ConsoleDefaults.USER_CONFIG_FILE", user_cfg_filename)

            with open(project_cfg_filename, "w") as project_f:
                project_f.write("\n".join(project_cfg))

            with open(user_cfg_filename, "w") as user_f:
                user_f.write("\n".join(user_cfg))

            # command-line options should override user_cfg and project_cfg
            args_dict = parse_args(["--parameters", "PARAMETERS-commandline"])

            assert args_dict["cluster_file"] == "CLUSTERFILE-project"
            assert args_dict["results_root"] == "RESULTSROOT-user"
            assert args_dict["parameters"] == "PARAMETERS-commandline"

        finally:
            shutil.rmtree(tmpdir)

    def check_config_file_option(self):
        """Check that config file option works"""
        tmpdir = tempfile.mkdtemp(dir="/tmp")
        user_cfg_filename = os.path.join(tmpdir, "ducktape-user.cfg")

        user_cfg = [
            "--results-root RESULTSROOT-user",
            "--parameters PARAMETERS-user"
        ]

        try:
            with open(user_cfg_filename, "w") as user_f:
                user_f.write("\n".join(user_cfg))
            args_dict = parse_args(["--config-file", user_cfg_filename])
            assert args_dict["results_root"] == "RESULTSROOT-user"
            assert args_dict["parameters"] == "PARAMETERS-user"
        finally:
            shutil.rmtree(tmpdir)



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from ducktape.command_line.defaults import ConsoleDefaults
from ducktape.utils.util import ducktape_version

import argparse
import itertools
import os
import sys


def create_ducktape_parser():
    parser = argparse.ArgumentParser(description="Discover and run your tests")
    parser.add_argument('test_path', metavar='test_path', type=str, nargs='*', default=[os.getcwd()],
                        help='One or more test identifiers or test suite paths to execute')
    parser.add_argument('--exclude', type=str, nargs='*', default=None,
                        help='one or more space-delimited strings indicating which tests to exclude')
    parser.add_argument("--collect-only", action="store_true", help="display collected tests, but do not run.")
    parser.add_argument("--collect-num-nodes", action="store_true",
                        help="display total number of nodes requested by all tests, but do not run anything.")
    parser.add_argument("--debug", action="store_true", help="pipe more verbose test output to stdout.")
    parser.add_argument("--config-file", action="store", default=ConsoleDefaults.USER_CONFIG_FILE,
                        help="path to project-specific configuration file.")
    parser.add_argument("--compress", action="store_true", help="compress remote logs before collection.")
    parser.add_argument("--cluster", action="store", default=ConsoleDefaults.CLUSTER_TYPE,
                        help="cluster class to use to allocate nodes for tests.")
    parser.add_argument("--default-num-nodes", action="store", type=int, default=None,
                        help="Global hint for cluster usage. A test without the @cluster annotation will "
                        "default to this value for expected cluster usage.")
    parser.add_argument("--cluster-file", action="store", default=None,
                        help="path to a json file which provides information needed to initialize a json cluster. "
                             "The file is used to read/write cached cluster info if "
                             "cluster is ducktape.cluster.vagrant.VagrantCluster.")
    parser.add_argument("--results-root", action="store", default=ConsoleDefaults.RESULTS_ROOT_DIRECTORY,
                        help="path to custom root results directory. Running ducktape with this root "
                             "specified will result in new test results being stored in a subdirectory of "
                             "this root directory.")
    parser.add_argument("--exit-first", action="store_true", help="exit after first failure")
    parser.add_argument("--no-teardown", action="store_true",
                        help="don't kill running processes or remove log files when a test has finished running. "
                             "This is primarily useful for test developers who want to interact with running "
                             "services after a test has run.")
    parser.add_argument("--version", action="store_true", help="display version")
    parser.add_argument("--parameters", action="store",
                        help="inject these arguments into the specified test(s). Specify parameters as a JSON string.")
    parser.add_argument("--globals", action="store",
                        help="user-defined globals go here. "
                             "This can be a file containing a JSON object, or a string representing a JSON object.")
    parser.add_argument("--max-parallel", action="store", type=int, default=1,
                        help="Upper bound on number of tests run simultaneously.")
    parser.add_argument("--repeat", action="store", type=int, default=1,
                        help="Use this flag to repeat all discovered tests the given number of times.")
    parser.add_argument("--subsets", action="store", type=int, default=1,
                        help="Number of subsets of tests to statically break the tests into to allow for parallel "
                             "execution without coordination between test runner processes.")
    parser.add_argument("--subset", action="store", type=int, default=0,
                        help="Which subset of the tests to run, based on the breakdown using the parameter for "
                             "--subsets")
    parser.add_argument("--historical-report", action="store", type=str,
                        help="URL of a JSON report file containing stats from a previous test run. If specified, "
                             "this will be used when creating subsets of tests to divide evenly by total run time "
                             "instead of by number of tests.")
    parser.add_argument("--skip-nodes-allocation", action="store_true", help="Use this flag to skip allocating "
                        "nodes for services. Can be used when running specific tests on a running platform")
    parser.add_argument("--sample", action="store", type=int,
                        help="The size of a random test sample to run")
    parser.add_argument("--fail-bad-cluster-utilization", action="store_true",
                        help="Fail a test if the cluster node utilization does not match the cluster node usage.")
    parser.add_argument("--test-runner-timeout", action="store", type=int, default=1800000,
                        help="Amount of time in milliseconds between test communicating between the test runner"
                             " before a timeout error occurs. Default is 30 minutes")
    parser.add_argument("--ssh-checker-function", action="store", type=str, nargs="+",
                        help="Python module path(s) to a function that takes an exception and a remote account"
                        " that will be called when an ssh error occurs, this can give some "
                        "validation or better logging when an ssh error occurs. Specify any "
                        "number of module paths after this flag to be called."),
    parser.add_argument("--deflake", action="store", type=int, default=1,
                        help="the number of times a failed test should be ran in total (including its initial run) "
                             "to determine flakyness. When not present, deflake will not be used, "
                             "and a test will be marked as either passed or failed. "
                             "When enabled tests will be marked as flaky if it passes on any of the reruns")
    return parser


def get_user_config_file(args):
    """Helper function to get specified (or default) user config file.
    :return Filename which is the path to the config file.
    """
    parser = create_ducktape_parser()
    config_file = vars(parser.parse_args(args))["config_file"]
    assert config_file is not None
    return os.path.expanduser(config_file)


def config_file_to_args_list(config_file):
    """Parse in contents of config file, and return a list of command-line options parseable by the ducktape parser.

    Skip whitespace lines and comments (lines prefixed by "#")
    """
    if config_file is None:
        raise RuntimeError("config_file is None")

    # Read in configuration, but ignore empty lines and comments
    config_lines = [line for line in open(config_file).readlines()
                    if (len(line.strip()) > 0 and line.lstrip()[0] != '#')]

    return list(itertools.chain(*[line.split() for line in config_lines]))


def parse_args(args):
    """Parse in command-line and config file options.

    Command line arguments have the highest priority, then user configs specified in ~/.ducktape/config, and finally
    project configs specified in <ducktape_dir>/config.
    """

    parser = create_ducktape_parser()

    if len(args) == 0:
        # Show help if there are no arguments
        parser.print_help()
        sys.exit(0)

    # Collect arguments from project config file, user config file, and command line
    # later arguments supersede earlier arguments
    args_list = []

    project_config_file = ConsoleDefaults.PROJECT_CONFIG_FILE
    if os.path.exists(project_config_file):
        args_list.extend(config_file_to_args_list(project_config_file))

    user_config_file = get_user_config_file(args)
    if os.path.exists(user_config_file):
        args_list.extend(config_file_to_args_list(user_config_file))

    args_list.extend(args)
    parsed_args_dict = vars(parser.parse_args(args_list))

    if parsed_args_dict["version"]:
        print(ducktape_version())
        sys.exit(0)

    return parsed_args_dict



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.cluster.vagrant import VagrantCluster
from ducktape.services.service import Service
import json
import pickle
import os
import random
import pytest
from ducktape.cluster.remoteaccount import RemoteAccountError

from tests.runner.fake_remote_account import create_fake_remote_account

TWO_HOSTS = """Host worker1
  HostName 127.0.0.1
  User vagrant
  Port 2222
  UserKnownHostsFile /dev/null
  StrictHostKeyChecking no
  PasswordAuthentication no
  IdentityFile /Users/foo/ducktape.git/.vagrant/machines/worker1/virtualbox/private_key
  IdentitiesOnly yes
  LogLevel FATAL

Host worker2
  HostName 127.0.0.2
  User vagrant
  Port 2200
  UserKnownHostsFile /dev/null
  StrictHostKeyChecking no
  PasswordAuthentication no
  IdentityFile /Users/foo/ducktape.git/.vagrant/machines/worker2/virtualbox/private_key
  IdentitiesOnly yes
  LogLevel FATAL

"""


def make_vagrant_cluster(*args, **kwargs):
    return VagrantCluster(make_remote_account_func=create_fake_remote_account, *args, **kwargs)


class CheckVagrantCluster(object):

    def setup_method(self, _):
        # We roll our own tempfile name instead of using python tempfile module because
        # in some cases, we want self.cluster_file to be the name of a file which does not yet exist
        self.cluster_file = "cluster_file_temporary-%d.json" % random.randint(1, 2**63 - 1)
        if os.path.exists(self.cluster_file):
            os.remove(self.cluster_file)

    def teardown_method(self, _):
        if os.path.exists(self.cluster_file):
            os.remove(self.cluster_file)

    def _set_monkeypatch_attr(self, monkeypatch):
        monkeypatch.setattr("ducktape.cluster.vagrant.VagrantCluster._vagrant_ssh_config", lambda vc: (TWO_HOSTS, None))
        monkeypatch.setattr(
            "ducktape.cluster.linux_remoteaccount.LinuxRemoteAccount.fetch_externally_routable_ip",
            lambda vc: "127.0.0.1")

    def check_pickleable(self, monkeypatch):
        self._set_monkeypatch_attr(monkeypatch)
        cluster = VagrantCluster()
        pickle.dumps(cluster)

    def check_one_host_parsing(self, monkeypatch):
        """check the behavior of VagrantCluster when cluster_file is not specified. VagrantCluster should read
        cluster information from _vagrant_ssh_config().
        """
        self._set_monkeypatch_attr(monkeypatch)

        cluster = make_vagrant_cluster()
        assert len(cluster) == 2
        assert cluster.num_available_nodes() == 2
        node1, node2 = cluster.alloc(Service.setup_cluster_spec(num_nodes=2))

        assert node1.account.hostname == "worker1"
        assert node1.account.user == "vagrant"
        assert node1.account.ssh_hostname == '127.0.0.1'

        assert node2.account.hostname == "worker2"
        assert node2.account.user == "vagrant"
        assert node2.account.ssh_hostname == '127.0.0.2'

    def check_cluster_file_write(self, monkeypatch):
        """check the behavior of VagrantCluster when cluster_file is specified but the file doesn't exist.
        VagrantCluster should read cluster information from _vagrant_ssh_config() and write the information to
        cluster_file.
        """
        self._set_monkeypatch_attr(monkeypatch)
        assert not os.path.exists(self.cluster_file)

        cluster = make_vagrant_cluster(cluster_file=self.cluster_file)
        cluster_json_expected = {}
        nodes = [
            {
                "externally_routable_ip": node_account.externally_routable_ip,
                "ssh_config": {
                    "host": node_account.ssh_config.host,
                    "hostname": node_account.ssh_config.hostname,
                    "user": node_account.ssh_config.user,
                    "identityfile": node_account.ssh_config.identityfile,
                    "password": node_account.ssh_config.password,
                    "port": node_account.ssh_config.port,
                    "connecttimeout": None
                }
            }
            for node_account in cluster._available_accounts
        ]

        cluster_json_expected["nodes"] = nodes

        cluster_json_actual = json.load(open(os.path.abspath(self.cluster_file)))
        assert cluster_json_actual == cluster_json_expected

    def check_cluster_file_read(self, monkeypatch):
        """check the behavior of VagrantCluster when cluster_file is specified and the file exists.
        VagrantCluster should read cluster information from cluster_file.
        """
        self._set_monkeypatch_attr(monkeypatch)

        # To verify that VagrantCluster reads cluster information from the cluster_file, the
        # content in the file is intentionally made different from that returned by _vagrant_ssh_config().
        nodes_expected = []
        node1_expected = {
            "externally_routable_ip": "127.0.0.3",
            "ssh_config": {
                "host": "worker3",
                "hostname": "127.0.0.3",
                "user": "vagrant",
                "port": 2222,
                "password": "password",
                "identityfile": "/path/to/identfile3",
                "connecttimeout": None
            }
        }
        nodes_expected.append(node1_expected)

        node2_expected = {
            "externally_routable_ip": "127.0.0.2",
            "ssh_config": {
                "host": "worker2",
                "hostname": "127.0.0.2",
                "user": "vagrant",
                "port": 2223,
                "password": None,
                "identityfile": "/path/to/indentfile2",
                "connecttimeout": 10
            }
        }
        nodes_expected.append(node2_expected)

        cluster_json_expected = {}
        cluster_json_expected["nodes"] = nodes_expected
        json.dump(cluster_json_expected, open(self.cluster_file, 'w+'),
                  indent=2, separators=(',', ': '), sort_keys=True)

        # Load the cluster from the json file we just created
        cluster = make_vagrant_cluster(cluster_file=self.cluster_file)

        assert len(cluster) == 2
        assert cluster.num_available_nodes() == 2
        node2, node3 = cluster.alloc(Service.setup_cluster_spec(num_nodes=2))

        assert node3.account.hostname == "worker2"
        assert node3.account.user == "vagrant"
        assert node3.account.ssh_hostname == '127.0.0.2'
        assert node3.account.ssh_config.to_json() == node2_expected["ssh_config"]

        assert node2.account.hostname == "worker3"
        assert node2.account.user == "vagrant"
        assert node2.account.ssh_hostname == '127.0.0.3'
        assert node2.account.ssh_config.to_json() == node1_expected["ssh_config"]

    def check_no_valid_network_devices(self, monkeypatch):
        """
        test to make sure that a remote account error is raised when no network devices are found
        """
        monkeypatch.setattr("ducktape.cluster.vagrant.VagrantCluster._vagrant_ssh_config", lambda vc: (TWO_HOSTS, None))
        monkeypatch.setattr("ducktape.cluster.linux_remoteaccount.LinuxRemoteAccount.get_network_devices",
                            lambda account: [])

        with pytest.raises(RemoteAccountError):
            VagrantCluster()



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.tests.test import Test
from ducktape.services.service import Service
from ducktape.mark import matrix


MEMORY_EATER_LIST_SIZE = 10000000
N_TEST_CASES = 5


class MemoryEater(Service):
    """Simple service that has a reference to a list with many elements"""

    def __init__(self, context):
        super(MemoryEater, self).__init__(context, 1)
        self.items = []

    def start_node(self, node):
        self.items = [x for x in range(MEMORY_EATER_LIST_SIZE)]

    def stop_node(self, node):
        pass

    def clean_node(self, node):
        pass

    @property
    def num_nodes(self):
        return 1


class MemoryLeakTest(Test):
    """A group of identical "memory-hungry" ducktape tests.
    Each test holds a reference to a service which itself holds a reference to a large (memory intensive) object.
    """

    def __init__(self, test_context):
        super(MemoryLeakTest, self).__init__(test_context)
        self.memory_eater = MemoryEater(test_context)

    @matrix(x=[i for i in range(N_TEST_CASES)])
    def test_leak(self, x):
        self.memory_eater.start()



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.tests.test import Test

NUM_TESTS = 3


class TestB(Test):
    """Loader should discover this."""

    def test_b(self):
        pass


class TestBB(Test):
    """Loader should discover this with 2 tests."""
    test_not_callable = 3

    def test_bb_one(self):
        pass

    def bb_two_test(self):
        pass

    def other_method(self):
        pass


class TestInvisible(object):
    """Loader should not discover this."""

    def test_invisible(self):
        pass



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import time

from ducktape.tests.test import TestContext
from ducktape.json_serializable import DucktapeJSONEncoder
from ducktape.tests.reporter import SingleResultFileReporter
from ducktape.utils.local_filesystem_utils import mkdir_p
from ducktape.utils.util import ducktape_version
from ducktape.tests.status import FLAKY, PASS, FAIL, IGNORE


class TestResult(object):
    """Wrapper class for a single result returned by a single test."""

    def __init__(self,
                 test_context,
                 test_index,
                 session_context,
                 test_status=PASS,
                 summary="",
                 data=None,
                 start_time=-1,
                 stop_time=-1):
        """
        @param test_context  standard test context object
        @param test_status   did the test pass or fail, etc?
        @param summary       summary information
        @param data          data returned by the test, e.g. throughput
        """
        self.nodes_allocated = len(test_context.cluster)
        if hasattr(test_context, "services"):
            self.services = test_context.services.to_json()
            self.nodes_used = test_context.services.min_cluster_spec().size()
        else:
            self.services = {}
            self.nodes_used = 0

        self.test_id = test_context.test_id
        self.module_name = test_context.module_name
        self.cls_name = test_context.cls_name
        self.function_name = test_context.function_name
        self.injected_args = test_context.injected_args
        self.description = test_context.description
        self.results_dir = TestContext.results_dir(test_context, test_index)

        self.test_index = test_index

        self.session_context = session_context
        self.test_status = test_status
        self.summary = summary
        self.data = data
        self.file_name = test_context.file

        self.base_results_dir = session_context.results_dir
        if not self.results_dir.endswith(os.path.sep):
            self.results_dir += os.path.sep
        if not self.base_results_dir.endswith(os.path.sep):
            self.base_results_dir += os.path.sep
        assert self.results_dir.startswith(self.base_results_dir)
        self.relative_results_dir = self.results_dir[len(self.base_results_dir):]

        # For tracking run time
        self.start_time = start_time
        self.stop_time = stop_time

    def __repr__(self):
        return "<%s - test_status:%s, data:%s>" % (self.__class__.__name__, self.test_status, str(self.data))

    @property
    def run_time_seconds(self):
        if self.start_time < 0:
            return -1
        if self.stop_time < 0:
            return time.time() - self.start_time

        return self.stop_time - self.start_time

    def report(self):
        if not os.path.exists(self.results_dir):
            mkdir_p(self.results_dir)

        self.dump_json()
        test_reporter = SingleResultFileReporter(self)
        test_reporter.report()

    def dump_json(self):
        """Dump this object as json to the given location. By default, dump into self.results_dir/report.json"""
        with open(os.path.join(self.results_dir, "report.json"), "w") as fd:
            json.dump(self, fd, cls=DucktapeJSONEncoder, sort_keys=True, indent=2)

    def to_json(self):
        return {
            "test_id": self.test_id,
            "module_name": self.module_name,
            "cls_name": self.cls_name,
            "function_name": self.function_name,
            "injected_args": self.injected_args,
            "description": self.description,
            "results_dir": self.results_dir,
            "relative_results_dir": self.relative_results_dir,
            "base_results_dir": self.base_results_dir,
            "test_status": self.test_status,
            "summary": self.summary,
            "data": self.data,
            "start_time": self.start_time,
            "stop_time": self.stop_time,
            "run_time_seconds": self.run_time_seconds,
            "nodes_allocated": self.nodes_allocated,
            "nodes_used": self.nodes_used,
            "services": self.services
        }


class TestResults(object):
    """Class used to aggregate individual TestResult objects from many tests."""

    def __init__(self, session_context, cluster):
        """
        :type session_context: ducktape.tests.session.SessionContext
        """
        self._results = []
        self.session_context = session_context
        self.cluster = cluster

        # For tracking total run time
        self.start_time = -1
        self.stop_time = -1

    def append(self, obj):
        return self._results.append(obj)

    def __len__(self):
        return len(self._results)

    def __iter__(self):
        return iter(self._results)

    @property
    def num_passed(self):
        return len([r for r in self._results if r.test_status == PASS])

    @property
    def num_failed(self):
        return len([r for r in self._results if r.test_status == FAIL])

    @property
    def num_ignored(self):
        return len([r for r in self._results if r.test_status == IGNORE])

    @property
    def num_flaky(self):
        return len([r for r in self._results if r.test_status == FLAKY])

    @property
    def run_time_seconds(self):
        if self.start_time < 0:
            return -1
        if self.stop_time < 0:
            self.stop_time = time.time()

        return self.stop_time - self.start_time

    def get_aggregate_success(self):
        """Check cumulative success of all tests run so far
        :rtype: bool
        """
        for result in self._results:
            if result.test_status == FAIL:
                return False
        return True

    def _stats(self, num_list):
        if len(num_list) == 0:
            return {
                "mean": None,
                "min": None,
                "max": None
            }

        return {
            "mean": sum(num_list) / float(len(num_list)),
            "min": min(num_list),
            "max": max(num_list)
        }

    def to_json(self):
        if self.run_time_seconds == 0 or len(self.cluster) == 0:
            # If things go horribly wrong, the test run may be effectively instantaneous
            # Let's handle this case gracefully, and avoid divide-by-zero
            cluster_utilization = 0
            parallelism = 0
        else:
            cluster_utilization = (1.0 / len(self.cluster)) * (1.0 / self.run_time_seconds) * \
                sum([r.nodes_used * r.run_time_seconds for r in self])
            parallelism = sum([r.run_time_seconds for r in self._results]) / self.run_time_seconds
        result = {
            "ducktape_version": ducktape_version(),
            "session_context": self.session_context,
            "run_time_seconds": self.run_time_seconds,
            "start_time": self.start_time,
            "stop_time": self.stop_time,
            "run_time_statistics": self._stats([r.run_time_seconds for r in self]),
            "cluster_nodes_used": self._stats([r.nodes_used for r in self]),
            "cluster_nodes_allocated": self._stats([r.nodes_allocated for r in self]),
            "cluster_utilization": cluster_utilization,
            "cluster_num_nodes": len(self.cluster),
            "num_passed": self.num_passed,
            "num_failed": self.num_failed,
            "num_ignored": self.num_ignored,
            "parallelism": parallelism,
            "results": [r for r in self._results]
        }
        if self.num_flaky:
            result['num_flaky'] = self.num_flaky
        return result



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.tests.runner import TestRunner
from ducktape.mark.mark_expander import MarkedFunctionExpander
from ducktape.cluster.localhost import LocalhostCluster

from .resources.test_memory_leak import MemoryLeakTest

import math
from memory_profiler import memory_usage
import os
from six.moves import queue
import statistics
from statistics import mean

import tests.ducktape_mock

from mock import Mock


N_TEST_CASES = 5


MEMORY_LEAK_TEST_FILE = os.path.abspath(
    os.path.join(
        os.path.dirname(__file__),
        "resources/test_memory_leak.py"
    )
)


class InstrumentedTestRunner(TestRunner):
    """Identical to TestRunner, except dump memory used by the current process
    before running each test.
    """

    def __init__(self, *args, **kwargs):
        self.queue = kwargs.get("queue")
        del kwargs["queue"]
        super(InstrumentedTestRunner, self).__init__(*args, **kwargs)

    def _run_single_test(self, test_context):
        # write current memory usage to file before running the test
        pid = os.getpid()
        current_memory = memory_usage(pid)[0]
        self.queue.put(current_memory)

        super(InstrumentedTestRunner, self)._run_single_test(test_context)


class CheckMemoryUsage(object):
    def setup_method(self, _):
        self.cluster = LocalhostCluster(num_nodes=100)
        self.session_context = tests.ducktape_mock.session_context()

    def check_for_inter_test_memory_leak(self):
        """Until v0.3.10, ducktape had a serious source of potential memory leaks.

        Because test_context objects held a reference to all services for the duration of a test run, the memory
        used by any individual service would not be garbage-collected until well after *all* tests had run.

        This memory leak was discovered in Kafka system tests, where many long-running system tests were enough
        to cumulatively use up the memory on the test machine, causing a cascade of test failures due to
        inability to allocate any more memory.

        This test provides a regression check against this type of memory leak; it fails without the fix, and passes
        with it.
        """
        # Get a list of test_context objects for the test runner
        ctx_list = []
        test_methods = [MemoryLeakTest.test_leak]
        for f in test_methods:
            ctx_list.extend(MarkedFunctionExpander(session_context=self.session_context, cls=MemoryLeakTest, function=f,
                                                   file=MEMORY_LEAK_TEST_FILE, cluster=self.cluster).expand())
        assert len(ctx_list) == N_TEST_CASES  # Sanity check

        q = queue.Queue()
        runner = InstrumentedTestRunner(self.cluster, self.session_context, Mock(), ctx_list, 1, queue=q)
        runner.run_all_tests()

        measurements = []
        while not q.empty():
            measurements.append(q.get())
        self.validate_memory_measurements(measurements)

    def validate_memory_measurements(self, measurements):
        """A rough heuristic to check that stair-case style memory leak is not present.

        The idea is that when well-behaved, in this specific test, the maximum memory usage should be near the "middle".
         Here we check that the maximum usage is within 5% of the median memory usage.

        What is meant by stair-case? When the leak was present in its most blatant form,
         each repetition of MemoryLeak.test_leak run by the test runner adds approximately a fixed amount of memory
         without freeing much, resulting in a memory usage profile that looks like a staircase going up to the right.
        """
        median_usage = statistics.median(measurements)
        max_usage = max(measurements)

        usage_stats = "\nmax: %s,\nmedian: %s,\nall: %s\n" % (max_usage, median_usage, measurements)

        # we want to make sure that max usage doesn't exceed median usage by very much
        relative_diff = (max_usage - median_usage) / median_usage
        slope = self._linear_regression_slope(measurements)

        if slope > 0:
            # check max memory usage iff the memory measurements seem to be increasing overall
            assert relative_diff <= .05, "max usage exceeded median usage by too much; there may " \
                                         "be a memory leak: %s" % usage_stats

    def _linear_regression_slope(self, arr):
        """Return the sign of the slope of the least squares fit line.
        """
        assert len(arr) > 0

        x_vals = [i for i in range(len(arr))]
        mean_x = mean(x_vals)
        mean_y = mean(arr)

        #            mean([x_i * y_i]) - mean_x * mean_y
        # slope =    -----------------------------------
        #                       variance([x_i])
        #
        # where variance is (1/N) * sum([(x_i - mean_x)^2])
        #
        # the denominator in regression formula is always positive, so it's enough to compute the numerator

        slope_numerator = mean([i * arr[i] for i in x_vals])
        slope_numerator = slope_numerator - (mean_x * mean_y)

        # return the sign
        return math.copysign(slope_numerator, 1)



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class TestScheduler(object):
    """This class tracks tests which are scheduled to run, and provides an ordering based on the current cluster state.

    The ordering is "on-demand"; calling next returns the largest cluster user which fits in the currently
    available cluster nodes.
    """

    def __init__(self, test_contexts, cluster):
        self.cluster = cluster

        # Track tests which would never be offered up by the scheduling algorithm due to insufficient
        # cluster resources
        self._test_context_list = test_contexts.copy()

        self._sort_test_context_list()

    def __len__(self):
        """Number of tests currently in the scheduler"""
        return len(self._test_context_list)

    def __iter__(self):
        return self

    def filter_unschedulable_tests(self):
        """
        Filter out tests that cannot be scheduled with the current cluster, remove them from
        this scheduler and return them.
        """
        all = self.cluster.all()
        unschedulable = []
        for test_context in self._test_context_list:
            if not all.nodes.can_remove_spec(test_context.expected_cluster_spec):
                unschedulable.append(test_context)
        for u in unschedulable:
            self._test_context_list.remove(u)
        return unschedulable

    def _sort_test_context_list(self):
        """Replace self.test_context_list with a sorted shallow copy

        Sort from largest cluster users to smallest
        """
        # sort from largest cluster users to smallest
        self._test_context_list = sorted(self._test_context_list,
                                         key=lambda tc: tc.expected_num_nodes,
                                         reverse=True)

    def peek(self):
        """Locate and return the next object to be scheduled, without removing it internally.

        :return test_context for the next test to be scheduled.
            If scheduler is empty, or no test can currently be scheduled, return None.
        """
        for tc in self._test_context_list:
            if self.cluster.available().nodes.can_remove_spec(tc.expected_cluster_spec):
                return tc

        return None

    def remove(self, tc):
        """Remove test context object from this scheduler.
        Intended usage is to peek() first, then perform whatever validity checks,
        and if they pass, remove() it from the scheduler.
        """
        if tc:
            self._test_context_list.remove(tc)



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import patch

import pytest

from ducktape.cluster.node_container import NodeContainer, InsufficientResourcesError
from ducktape.tests.runner_client import RunnerClient
from ducktape.tests.status import PASS, FAIL
from ducktape.tests.test import TestContext
from ducktape.tests.runner import TestRunner
from ducktape.mark.mark_expander import MarkedFunctionExpander
from ducktape.cluster.localhost import LocalhostCluster
from tests.ducktape_mock import FakeCluster

import tests.ducktape_mock
from tests.runner.resources.test_fails_to_init import FailsToInitTest
from tests.runner.resources.test_fails_to_init_in_setup import FailsToInitInSetupTest
from .resources.test_bad_actor import BadActorTest
from .resources.test_thingy import ClusterTestThingy, TestThingy
from .resources.test_failing_tests import FailingTest
from ducktape.tests.reporter import JUnitReporter
from ducktape.errors import TimeoutError

from mock import Mock
import os
import xml.etree.ElementTree as ET

from .resources.test_various_num_nodes import VariousNumNodesTest

TEST_THINGY_FILE = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "resources/test_thingy.py"))
FAILING_TEST_FILE = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "resources/test_failing_tests.py"))
FAILS_TO_INIT_TEST_FILE = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "resources/test_fails_to_init.py"))
FAILS_TO_INIT_IN_SETUP_TEST_FILE = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "resources/test_fails_to_init_in_setup.py"))
VARIOUS_NUM_NODES_TEST_FILE = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "resources/test_various_num_nodes.py"))
BAD_ACTOR_TEST_FILE = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "resources/test_bad_actor.py"))


class CheckRunner(object):
    def check_insufficient_cluster_resources(self):
        """The test runner should behave sensibly when the cluster is too small to run a given test."""
        mock_cluster = FakeCluster(1)
        session_context = tests.ducktape_mock.session_context()

        test_context = TestContext(session_context=session_context, module=None, cls=TestThingy,
                                   function=TestThingy.test_pi, file=TEST_THINGY_FILE, cluster=mock_cluster)
        runner = TestRunner(mock_cluster, session_context, Mock(), [test_context], 1)

        # Even though the cluster is too small, the test runner should this handle gracefully without raising an error
        results = runner.run_all_tests()
        assert len(results) == 1
        assert results.num_failed == 1
        assert results.num_passed == 0
        assert results.num_ignored == 0

    def _do_expand(self, test_file, test_class, test_methods, cluster=None, session_context=None):
        ctx_list = []
        for f in test_methods:
            ctx_list.extend(
                MarkedFunctionExpander(
                    session_context=session_context,
                    cls=test_class, function=f, file=test_file, cluster=cluster).expand())
        return ctx_list

    def check_simple_run(self):
        """Check expected behavior when running a single test."""
        mock_cluster = LocalhostCluster(num_nodes=1000)
        session_context = tests.ducktape_mock.session_context()

        test_methods = [TestThingy.test_pi, TestThingy.test_ignore1, TestThingy.test_ignore2, TestThingy.test_failure]
        ctx_list = self._do_expand(test_file=TEST_THINGY_FILE, test_class=TestThingy, test_methods=test_methods,
                                   cluster=mock_cluster, session_context=session_context)
        runner = TestRunner(mock_cluster, session_context, Mock(), ctx_list, 1)

        results = runner.run_all_tests()
        assert len(results) == 4
        assert results.num_flaky == 0
        assert results.num_failed == 1
        assert results.num_passed == 1
        assert results.num_ignored == 2

        result_with_data = [r for r in results if r.data is not None][0]
        assert result_with_data.data == {"data": 3.14159}

    def check_deflake_run(self):
        """Check expected behavior when running a single test."""
        mock_cluster = LocalhostCluster(num_nodes=1000)
        session_context = tests.ducktape_mock.session_context()

        test_methods = [TestThingy.test_flaky, TestThingy.test_failure]
        ctx_list = self._do_expand(test_file=TEST_THINGY_FILE, test_class=TestThingy, test_methods=test_methods,
                                   cluster=mock_cluster, session_context=session_context)
        runner = TestRunner(mock_cluster, session_context, Mock(), ctx_list, 2)

        results = runner.run_all_tests()
        assert len(results) == 2
        assert results.num_flaky == 1
        assert results.num_failed == 1
        assert results.num_passed == 0
        assert results.num_ignored == 0

    def check_runner_report_junit(self):
        """Check we can serialize results into a xunit xml format. Also ensures that the XML report
        adheres to the Junit spec using xpath queries"""
        mock_cluster = LocalhostCluster(num_nodes=1000)
        session_context = tests.ducktape_mock.session_context()
        test_methods = [TestThingy.test_pi, TestThingy.test_ignore1, TestThingy.test_ignore2, TestThingy.test_failure]
        ctx_list = self._do_expand(test_file=TEST_THINGY_FILE, test_class=TestThingy, test_methods=test_methods,
                                   cluster=mock_cluster, session_context=session_context)
        runner = TestRunner(mock_cluster, session_context, Mock(), ctx_list, 1)

        results = runner.run_all_tests()
        JUnitReporter(results).report()
        xml_report = os.path.join(session_context.results_dir, "report.xml")
        assert os.path.exists(xml_report)
        tree = ET.parse(xml_report)
        assert len(tree.findall('./testsuite/testcase/failure')) == 1
        assert len(tree.findall('./testsuite/testcase/skipped')) == 2
        assert len(tree.findall('./testsuite/testcase')) == 4

        passed = tree.findall("./testsuite/testcase/[@status='pass']")
        assert len(passed) == 1
        assert passed[0].get("classname") == "TestThingy"
        assert passed[0].get("name") == "test_pi"

        failures = tree.findall("./testsuite/testcase/[@status='fail']")
        assert len(failures) == 1
        assert failures[0].get("classname") == "TestThingy"
        assert failures[0].get("name") == "test_failure"

        ignores = tree.findall("./testsuite/testcase/[@status='ignore']")
        assert len(ignores) == 2
        assert ignores[0].get("classname") == "TestThingy"
        assert ignores[1].get("classname") == "TestThingy"

        assert ignores[0].get("name") == "test_ignore1"
        assert ignores[1].get("name") == "test_ignore2.x=5"

    def check_exit_first(self):
        """Confirm that exit_first in session context has desired effect of preventing any tests from running
        after the first test failure.
        """
        mock_cluster = LocalhostCluster(num_nodes=1000)
        session_context = tests.ducktape_mock.session_context(**{"exit_first": True})

        test_methods = [FailingTest.test_fail]
        ctx_list = self._do_expand(test_file=FAILING_TEST_FILE, test_class=FailingTest, test_methods=test_methods,
                                   cluster=mock_cluster, session_context=session_context)
        runner = TestRunner(mock_cluster, session_context, Mock(), ctx_list, 1)
        results = runner.run_all_tests()
        assert len(ctx_list) > 1
        assert len(results) == 1

    def check_exits_if_failed_to_initialize(self):
        """Validate that runner exits correctly when tests failed to initialize.
        """
        mock_cluster = LocalhostCluster(num_nodes=1000)
        session_context = tests.ducktape_mock.session_context()

        ctx_list = self._do_expand(test_file=FAILS_TO_INIT_TEST_FILE, test_class=FailsToInitTest,
                                   test_methods=[FailsToInitTest.test_nothing],
                                   cluster=mock_cluster, session_context=session_context)
        ctx_list.extend(self._do_expand(test_file=FAILS_TO_INIT_IN_SETUP_TEST_FILE, test_class=FailsToInitInSetupTest,
                                        test_methods=[FailsToInitInSetupTest.test_nothing],
                                        cluster=mock_cluster, session_context=session_context))

        runner = TestRunner(mock_cluster, session_context, Mock(), ctx_list, 1)
        results = runner.run_all_tests()
        # These tests fail to initialize, each class has two test methods, so should have 4 results, all failed
        assert len(results) == 4
        assert results.num_flaky == 0
        assert results.num_failed == 4
        assert results.num_passed == 0
        assert results.num_ignored == 0

    # mock an error reporting test failure - this should not prevent subsequent tests from execution and mark
    # failed test as failed correctly
    @patch.object(RunnerClient, '_exc_msg', side_effect=Exception)
    def check_sends_result_when_error_reporting_exception(self, exc_msg_mock):
        """Validates that an error when reporting an exception in the test doesn't prevent subsequent tests
        from executing"""
        mock_cluster = LocalhostCluster(num_nodes=1000)
        session_context = tests.ducktape_mock.session_context()
        test_methods = [TestThingy.test_failure, TestThingy.test_pi]
        ctx_list = self._do_expand(test_file=TEST_THINGY_FILE, test_class=TestThingy, test_methods=test_methods,
                                   cluster=mock_cluster, session_context=session_context)
        runner = TestRunner(mock_cluster, session_context, Mock(), ctx_list, 1)

        results = runner.run_all_tests()
        assert len(results) == 2
        assert results.num_flaky == 0
        assert results.num_failed == 1
        assert results.num_passed == 1
        assert results.num_ignored == 0

    def check_run_failure_with_bad_cluster_allocation(self):
        """Check test should be marked failed if it under-utilizes the cluster resources."""
        mock_cluster = LocalhostCluster(num_nodes=1000)
        session_context = tests.ducktape_mock.session_context(
            fail_bad_cluster_utilization="fail_bad_cluster_utilization")

        test_methods = [ClusterTestThingy.test_bad_num_nodes, ClusterTestThingy.test_good_num_nodes]
        ctx_list = self._do_expand(test_file=TEST_THINGY_FILE, test_class=ClusterTestThingy,
                                   test_methods=test_methods, cluster=mock_cluster,
                                   session_context=session_context)
        runner = TestRunner(mock_cluster, session_context, Mock(), ctx_list, 1)

        results = runner.run_all_tests()

        assert len(results) == 2
        assert results.num_flaky == 0
        assert results.num_failed == 1
        assert results.num_passed == 1
        assert results.num_ignored == 0

    def check_test_failure_with_too_many_nodes_requested(self):
        mock_cluster = LocalhostCluster(num_nodes=1000)
        session_context = tests.ducktape_mock.session_context(debug=True)

        ctx_list = self._do_expand(test_file=BAD_ACTOR_TEST_FILE, test_class=BadActorTest,
                                   test_methods=[BadActorTest.test_too_many_nodes],
                                   cluster=mock_cluster, session_context=session_context)
        ctx_list.extend(self._do_expand(test_file=VARIOUS_NUM_NODES_TEST_FILE, test_class=VariousNumNodesTest,
                                        test_methods=[VariousNumNodesTest.test_one_node_a],
                                        cluster=mock_cluster, session_context=session_context))
        runner = TestRunner(mock_cluster, session_context, Mock(), ctx_list, 1)
        results = runner.run_all_tests()
        assert results.num_flaky == 0
        assert results.num_failed == 1
        assert results.num_passed == 1
        assert results.num_ignored == 0
        passed = [r for r in results if r.test_status == PASS]
        failed = [r for r in results if r.test_status == FAIL]
        assert passed[0].test_id == 'tests.runner.resources.test_various_num_nodes.VariousNumNodesTest.test_one_node_a'
        assert failed[0].test_id == 'tests.runner.resources.test_bad_actor.BadActorTest.test_too_many_nodes'

    def check_runner_timeout(self):
        """Check process cleanup and error handling in a parallel runner client run."""
        mock_cluster = LocalhostCluster(num_nodes=1000)
        session_context = tests.ducktape_mock.session_context(max_parallel=1000, test_runner_timeout=1)

        test_methods = [TestThingy.test_delayed, TestThingy.test_failure]
        ctx_list = self._do_expand(test_file=TEST_THINGY_FILE, test_class=TestThingy, test_methods=test_methods,
                                   cluster=mock_cluster, session_context=session_context)
        runner = TestRunner(mock_cluster, session_context, Mock(), ctx_list, 1)

        with pytest.raises(TimeoutError):
            runner.run_all_tests()

        assert not runner._client_procs

    def check_cluster_shrink(self):
        """
        Check what happens if cluster loses a node while the runner is already running.
        SchedulerTestThingy has two 5-node tests, and one of each for 4, 3, and 2 nodes.

        Thus both 5-node tests should pass, first one failing during pre-allocation phase,
        second one shouldn't even attempt to be allocated.
        And all the other tests should pass still.
        """

        mock_cluster = ShrinkingLocalhostCluster(num_nodes=5)
        session_context = tests.ducktape_mock.session_context(max_parallel=10)

        test_methods = [
            VariousNumNodesTest.test_five_nodes_a,
            VariousNumNodesTest.test_five_nodes_b,
            VariousNumNodesTest.test_four_nodes,
            VariousNumNodesTest.test_three_nodes_a,
            VariousNumNodesTest.test_two_nodes_a
        ]

        ctx_list = self._do_expand(test_file=VARIOUS_NUM_NODES_TEST_FILE, test_class=VariousNumNodesTest,
                                   test_methods=test_methods, cluster=mock_cluster,
                                   session_context=session_context)

        runner = TestRunner(mock_cluster, session_context, Mock(), ctx_list, 1)
        results = runner.run_all_tests()

        assert len(results) == 5
        assert results.num_flaky == 0
        assert results.num_failed == 2  # both of the 5-node tests should fail
        assert results.num_passed == 3  # 4-node, 3-node and 2-node should all pass
        assert results.num_ignored == 0

    def check_cluster_shrink_reschedule(self):
        """
        Test that the test that failed to schedule initially due to a node going offline is not lost and is still
        scheduled when more nodes become available.

        We start with a 6-node cluster.
        First we run a long-ish 3-node test, leaving 3 nodes available.
        Then when trying to run a second 3-node test, we shrink the cluster, emulating one of the nodes
        going down - this leaves only 2 nodes available, so we cannot run this test.

        However, after the first 3-node test finishes running, it will return its 3 nodes back to the cluster,
        so the second 3-node test becomes schedulable again - this is what we test for.

        Also two two-node tests should pass too - they should be scheduled before the second 3-node test,
        while two nodes are waiting for the first 3-node test to finish.

        It's generally not a good practice to rely on sleep, but I think it's acceptable in this case,
        since we do need to rely on parallelism.
        """

        mock_cluster = ShrinkingLocalhostCluster(num_nodes=6, shrink_on=2)
        session_context = tests.ducktape_mock.session_context(max_parallel=10)

        test_methods = [
            VariousNumNodesTest.test_three_nodes_asleep,
            VariousNumNodesTest.test_three_nodes_b,
            VariousNumNodesTest.test_two_nodes_a,
            VariousNumNodesTest.test_two_nodes_b
        ]

        ctx_list = self._do_expand(test_file=VARIOUS_NUM_NODES_TEST_FILE, test_class=VariousNumNodesTest,
                                   test_methods=test_methods, cluster=mock_cluster,
                                   session_context=session_context)

        runner = TestRunner(mock_cluster, session_context, Mock(), ctx_list, 1)
        results = runner.run_all_tests()

        assert len(results) == 4
        assert results.num_flaky == 0
        assert results.num_failed == 0
        assert results.num_passed == 4
        assert results.num_ignored == 0

        # normal order on a 6-node cluster would be:
        #  - test_three_nodes_asleep, test_three_nodes_b, test_two_nodes_a, test_two_nodes_b
        # however the cluster would shrink to 5 nodes after scheduling the first 3-node test,
        # leaving no space for the second 3-node test to be scheduled, bumping it down the line,
        # while two 2-node tests will be scheduled alongside the
        expected_scheduling_order = [
            "VariousNumNodesTest.test_three_nodes_asleep",
            "VariousNumNodesTest.test_two_nodes_a",
            "VariousNumNodesTest.test_two_nodes_b",
            "VariousNumNodesTest.test_three_nodes_b"
        ]
        # We check the actual order the tests were scheduled in, since completion order might be different,
        # with so many fast tests running in parallel.
        actual_scheduling_order = [x.test_id for x in runner.test_schedule_log]
        assert actual_scheduling_order == expected_scheduling_order

    def check_cluster_shrink_to_zero(self):
        """
        Validates that if the cluster is shrunk to zero nodes size, no tests can run,
        but we still exit gracefully.
        """

        mock_cluster = ShrinkingLocalhostCluster(num_nodes=1, shrink_on=1)
        session_context = tests.ducktape_mock.session_context(max_parallel=10)

        test_methods = [
            VariousNumNodesTest.test_one_node_a,
            VariousNumNodesTest.test_one_node_b,
        ]

        ctx_list = self._do_expand(test_file=VARIOUS_NUM_NODES_TEST_FILE, test_class=VariousNumNodesTest,
                                   test_methods=test_methods, cluster=mock_cluster,
                                   session_context=session_context)

        runner = TestRunner(mock_cluster, session_context, Mock(), ctx_list, 1)
        results = runner.run_all_tests()

        assert len(results) == 2
        assert results.num_flaky == 0
        assert results.num_failed == 2
        assert results.num_passed == 0
        assert results.num_ignored == 0


class ShrinkingLocalhostCluster(LocalhostCluster):

    def __init__(self, *args, shrink_on=1, **kwargs):
        super().__init__(*args, **kwargs)
        self.bad_nodes = NodeContainer()
        # which call to shrink on
        self.shrink_on = shrink_on
        self.num_alloc_calls = 0

    def do_alloc(self, cluster_spec):
        allocated = super().do_alloc(cluster_spec)
        self.num_alloc_calls += 1
        if self.shrink_on == self.num_alloc_calls:
            bad_node = allocated.pop()
            self._in_use_nodes.remove_node(bad_node)
            self.bad_nodes.add_node(bad_node)

            # simplified logic, we know all nodes are of the same OS/type
            # check if we don't have enough nodes any more
            # (which really should be true every time, since the largest test would be scheduled)
            if len(allocated) < len(cluster_spec):
                # return all good nodes back to be available
                for node in allocated:
                    self._in_use_nodes.remove_node(node)
                    self._available_nodes.add_node(node)

                raise InsufficientResourcesError("yeah")

        return allocated



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.errors import TimeoutError
from tests.ducktape_mock import MockAccount
from tests.test_utils import find_available_port
from ducktape.cluster.remoteaccount import RemoteAccount
from ducktape.cluster.remoteaccount import RemoteAccountSSHConfig
import pytest

import logging
from threading import Thread
from six.moves import SimpleHTTPServer
from six.moves import socketserver
import threading
import time


class DummyException(Exception):
    pass


def raise_error_checker(error, remote_account):
    raise DummyException("dummy raise: {}\nfrom: {}".format(error, remote_account))


def raise_no_error_checker(error, remote_account):
    pass


class SimpleServer(object):
    """Helper class which starts a simple server listening on localhost at the specified port
    """

    def __init__(self):
        self.port = find_available_port()
        self.handler = SimpleHTTPServer.SimpleHTTPRequestHandler
        self.httpd = socketserver.TCPServer(("", self.port), self.handler)
        self.close_signal = threading.Event()
        self.server_started = False

    def start(self, delay_sec=0.0):
        """Run the server after specified delay"""
        def run():
            self.close_signal.wait(delay_sec)

            if not self.close_signal.is_set():
                self.server_started = True
                self.httpd.serve_forever()

        self.background_thread = Thread(target=run)
        self.background_thread.start()

    def stop(self):
        self.close_signal.set()

        if self.server_started:
            self.httpd.shutdown()
        self.background_thread.join(timeout=.5)
        if self.background_thread.is_alive():
            raise Exception("SimpleServer failed to stop quickly")


class CheckRemoteAccount(object):
    def setup(self):
        self.server = SimpleServer()
        self.account = MockAccount()

    def check_wait_for_http(self):
        """Check waiting without timeout"""
        self.server.start(delay_sec=0.0)
        self.account.wait_for_http_service(port=self.server.port, headers={}, timeout=10, path="/")

    def check_wait_for_http_timeout(self):
        """Check waiting with timeout"""

        timeout = 1
        start = time.time()
        self.server.start(delay_sec=5)

        try:
            self.account.wait_for_http_service(port=self.server.port, headers={}, timeout=timeout, path='/')
            raise Exception("Should have timed out waiting for server to start")
        except TimeoutError:
            # expected behavior. Now check that we're reasonably close to the expected timeout
            # This is a fairly loose check since there are various internal timeouts that can affect the overall
            # timing
            actual_timeout = time.time() - start
            assert abs(actual_timeout - timeout) / timeout < 1

    @pytest.mark.parametrize("checkers", [[raise_error_checker],
                                          [raise_no_error_checker, raise_error_checker],
                                          [raise_error_checker, raise_no_error_checker]])
    def check_ssh_checker(self, checkers):
        self.server.start()
        ssh_config = RemoteAccountSSHConfig.from_string(
            """
        Host dummy_host.com
            Hostname dummy_host.name.com
            Port 22
            User dummy
            ConnectTimeout 1
        """)
        self.account = RemoteAccount(ssh_config, ssh_exception_checks=checkers)
        with pytest.raises(DummyException):
            self.account.ssh('echo test')

    def teardown(self):
        self.server.stop()


class CheckRemoteAccountEquality(object):

    def check_remote_account_equality(self):
        """Different instances of remote account initialized with the same parameters should be equal."""

        ssh_config = RemoteAccountSSHConfig(host="thehost", hostname="localhost", port=22)

        kwargs = {
            "ssh_config": ssh_config,
            "externally_routable_ip": "345",
            "logger": logging.getLogger(__name__)
        }
        r1 = RemoteAccount(**kwargs)
        r2 = RemoteAccount(**kwargs)

        assert r1 == r2



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
import time


class ClientEventFactory(object):
    """Used by test runner clients to generate events."""

    READY = "READY"  # reply: {test_metadata, cluster, session_context}
    SETTING_UP = "SETTING_UP"
    RUNNING = "RUNNING"
    TEARING_DOWN = "TEARING_DOWN"
    FINISHED = "FINISHED"
    LOG = "LOG"

    # Types of messages available
    TYPES = {READY, SETTING_UP, RUNNING, TEARING_DOWN, FINISHED, LOG}

    def __init__(self, test_id, test_index, source_id):
        self.test_id = test_id
        # id of event source
        self.test_index = test_index
        self.source_id = source_id
        self.event_id = 0

    def _event(self, event_type, payload=None):
        """Create a message object with certain base fields, and augmented by the payload.

        :param event_type: type of message this is
        :param payload: a dict containing extra fields for the message. Key names should not conflict with keys
            in the base event.
        """
        assert event_type in ClientEventFactory.TYPES, "Unknown event type"
        if payload is None:
            payload = {}

        event = {
            "test_id": self.test_id,
            "source_id": self.source_id,
            "test_index": self.test_index,
            "event_id": self.event_id,
            "event_type": event_type,
            "event_time": time.time()
        }

        assert len(set(event.keys()).intersection(set(payload.keys()))) == 0, \
            "Payload and base event should not share keys. base event: %s, payload: %s" % (str(event), str(payload))

        event. update(payload)
        self.event_id += 1
        return event

    def copy(self, event):
        """Copy constructor: return a copy of the original message, but with a unique message id."""
        new_event = copy.copy(event)
        new_event["message_id"] = self.event_id
        self.event_id += 1

        return new_event

    def running(self):
        return self._event(
            event_type=ClientEventFactory.RUNNING,
            payload={
                "pid": os.getpid(),
                "pgroup_id": os.getpgrp()
            }
        )

    def ready(self):
        return self._event(
            event_type=ClientEventFactory.READY,
            payload={
                "pid": os.getpid(),
                "pgroup_id": os.getpgrp()
            }
        )

    def setting_up(self):
        return self._event(
            event_type=ClientEventFactory.SETTING_UP
        )

    def finished(self, result):
        return self._event(
            event_type=ClientEventFactory.FINISHED,
            payload={
                "result": result
            }
        )

    def log(self, message, level):
        return self._event(
            event_type=ClientEventFactory.LOG,
            payload={
                "message": message,
                "log_level": level
            }
        )


class EventResponseFactory(object):
    """Used by the test runner to create responses to events from client processes."""

    def _event_response(self, client_event, payload=None):
        if payload is None:
            payload = {}

        event_response = {
            "ack": True,
            "source_id": client_event["source_id"],
            "event_id": client_event["event_id"]
        }

        assert len(set(event_response.keys()).intersection(set(payload.keys()))) == 0, \
            "Payload and base event should not share keys. base event: %s, payload: %s" % (
                str(event_response), str(payload))

        event_response.update(payload)
        return event_response

    def running(self, client_event):
        return self._event_response(client_event)

    def ready(self, client_event, session_context, test_context, cluster):
        payload = {
            "session_context": session_context,
            "test_metadata": test_context.test_metadata,
            "cluster": cluster
        }

        return self._event_response(client_event, payload)

    def setting_up(self, client_event):
        return self._event_response(client_event)

    def finished(self, client_event):
        return self._event_response(client_event)

    def log(self, client_event):
        return self._event_response(client_event)



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.cluster.cluster_spec import ClusterSpec


class CheckClusterSpec(object):
    def check_cluster_spec_sizes(self):
        simple_linux_2 = ClusterSpec.simple_linux(2)
        assert 2 == len(simple_linux_2)
        assert 0 == len(ClusterSpec.empty())

    def check_to_string(self):
        empty = ClusterSpec.empty()
        assert "[]" == str(empty)
        simple_linux_5 = ClusterSpec.simple_linux(5)
        assert '[{"num_nodes": 5, "os": "linux"}]' == str(simple_linux_5)



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.tests.test import Test
from ducktape.mark import matrix

"""All tests in this module fail"""


class FailsToInitTest(Test):
    """All tests in this class fail to initialize due to an exception in constructor
    """
    def __init__(self, test_context):
        super(FailsToInitTest, self).__init__(test_context)
        x = None
        x.split(':')

    @matrix(x=[_ for _ in range(2)])
    def test_nothing(self):
        self.logger.info('NOTHING')



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.cluster.cluster import Cluster
from ducktape.cluster.cluster_spec import ClusterSpec, LINUX
from ducktape.cluster.node_container import NodeContainer
from ducktape.tests.session import SessionContext
from ducktape.tests.test import TestContext
from ducktape.cluster.linux_remoteaccount import LinuxRemoteAccount
from ducktape.cluster.remoteaccount import RemoteAccountSSHConfig
from mock import MagicMock


import os
import tempfile


def mock_cluster():
    return MagicMock()


class FakeClusterNode(object):
    @property
    def operating_system(self):
        return LINUX


class FakeCluster(Cluster):
    """A cluster class with counters, but no actual node objects"""

    def __init__(self, num_nodes):
        super(FakeCluster, self).__init__()
        self._available_nodes = NodeContainer()
        for i in range(0, num_nodes):
            self._available_nodes.add_node(FakeClusterNode())
        self._in_use_nodes = NodeContainer()

    def do_alloc(self, cluster_spec):
        good_nodes, bad_nodes = self._available_nodes.remove_spec(cluster_spec)
        self._in_use_nodes.add_nodes(good_nodes)
        return good_nodes

    def free_single(self, node):
        self._in_use_nodes.remove_node(node)
        self._available_nodes.add_node(node)

    def available(self):
        return ClusterSpec.from_nodes(self._available_nodes)

    def used(self):
        return ClusterSpec.from_nodes(self._in_use_nodes)


def session_context(**kwargs):
    """Return a SessionContext object"""

    if "results_dir" not in kwargs.keys():
        tmp = tempfile.mkdtemp()
        session_dir = os.path.join(tmp, "test_dir")
        os.mkdir(session_dir)
        kwargs["results_dir"] = session_dir

    return SessionContext(session_id="test_session", **kwargs)


def test_context(session_context=session_context(), cluster=mock_cluster()):
    """Return a TestContext object"""
    return TestContext(session_context=session_context, file="a/b/c", cluster=cluster)


class MockNode(object):
    """Mock cluster node"""

    def __init__(self):
        self.account = MockAccount()


class MockAccount(LinuxRemoteAccount):
    """Mock node.account object. It's Linux because tests are run in Linux."""

    def __init__(self, **kwargs):
        ssh_config = RemoteAccountSSHConfig(
            host="localhost",
            user=None,
            hostname="localhost",
            port=22)

        super(MockAccount, self).__init__(ssh_config, externally_routable_ip="localhost", logger=None, **kwargs)



# Copyright 2022 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ducktape.services.service import Service
from ducktape.tests.test import Test
from ducktape.mark.resource import cluster
import time


class SimpleEchoService(Service):
    """Simple service that allocates one node for performing tests of RemoteAccount functionality"""
    logs = {
        "my_log": {
            "path": "/tmp/log",
            "collect_default": True
        },
    }

    def __init__(self, context):
        super(SimpleEchoService, self).__init__(context, num_nodes=1)
        self.count = 0

    def echo(self):
        self.nodes[0].account.ssh("echo {} >> /tmp/log".format(self.count))
        self.count += 1


class SimpleRunnerTest(Test):
    def setup(self):
        self.service = SimpleEchoService(self.test_context)

    @cluster(num_nodes=1)
    def timeout_test(self):
        """
        a simple longer running test to test special run flags agaisnt.
        """
        self.service.start()

        while self.service.count < 100000000:
            self.service.echo()
            time.sleep(.2)



# Copyright 2014 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.cluster.cluster_spec import ClusterSpec
from ducktape.cluster.node_container import NodeContainer
from .cluster import Cluster, ClusterNode
from .linux_remoteaccount import LinuxRemoteAccount
from .remoteaccount import RemoteAccountSSHConfig


class LocalhostCluster(Cluster):
    """
    A "cluster" that runs entirely on localhost using default credentials. This doesn't require any user
    configuration and is equivalent to the old defaults in cluster_config.json. There are no constraints
    on the resources available.
    """

    def __init__(self, *args, **kwargs):
        super(LocalhostCluster, self).__init__()
        num_nodes = kwargs.get("num_nodes", 1000)
        self._available_nodes = NodeContainer()
        for i in range(num_nodes):
            ssh_config = RemoteAccountSSHConfig("localhost%d" % i, hostname="localhost", port=22)
            self._available_nodes.add_node(ClusterNode(
                LinuxRemoteAccount(ssh_config,
                                   ssh_exception_checks=kwargs.get("ssh_exception_checks"))))
        self._in_use_nodes = NodeContainer()

    def do_alloc(self, cluster_spec):
        # there shouldn't be any bad nodes in localhost cluster
        # since ClusterNode object does not implement `available()` method
        good_nodes, bad_nodes = self._available_nodes.remove_spec(cluster_spec)
        self._in_use_nodes.add_nodes(good_nodes)
        return good_nodes

    def free_single(self, node):
        self._in_use_nodes.remove_node(node)
        self._available_nodes.add_node(node)
        node.account.close()

    def available(self):
        return ClusterSpec.from_nodes(self._available_nodes)

    def used(self):
        return ClusterSpec.from_nodes(self._in_use_nodes)



"""
This module contains tests that are useful for developer debugging
and can contain sleep statements or test that intentionally fail or break things.
They're separate from test_remote_account.py for that reason.
"""
import time

from ducktape.mark import matrix, parametrize, ignore
from ducktape.mark.resource import cluster
from ducktape.tests.test import Test
from systests.cluster.test_remote_account import GenericService


class FailingTest(Test):
    """
    The purpose of this test is to validate reporters. Some of them are intended to fail.
    """
    def setup(self):
        self.service = GenericService(self.test_context, 1)

    @cluster(num_nodes=1)
    @matrix(string_param=['success-first', 'fail-second', 'fail-third'], int_param=[10, 20, -30])
    def matrix_test(self, string_param, int_param):
        assert not string_param.startswith('fail') and int_param > 0

    @cluster(num_nodes=1)
    @parametrize(string_param='success-first', int_param=10)
    @parametrize(string_param='fail-second', int_param=-10)
    def parametrized_test(self, string_param, int_param):
        assert not string_param.startswith('fail') and int_param > 0

    @cluster(num_nodes=1)
    def failing_test(self):
        assert False

    @cluster(num_nodes=1)
    def successful_test(self):
        assert True


class DebugThisTest(Test):

    @cluster(num_nodes=1)
    def one_node_test_sleep_90s(self):
        self.service = GenericService(self.test_context, 1)
        self.logger.warning('one_node_test - Sleeping for 90s')
        time.sleep(90)
        assert True

    @cluster(num_nodes=1)
    def one_node_test_sleep_30s(self):
        self.service = GenericService(self.test_context, 1)
        self.logger.warning('another_one_node_test - Sleeping for 30s')
        time.sleep(30)
        assert True

    @cluster(num_nodes=1)
    def another_one_node_test_sleep_30s(self):
        self.service = GenericService(self.test_context, 1)
        self.logger.warning('yet_another_one_node_test - Sleeping for 30s')
        time.sleep(30)
        assert True

    @cluster(num_nodes=2)
    def two_node_test(self):
        self.service = GenericService(self.test_context, 2)
        assert True

    @cluster(num_nodes=2)
    def another_two_node_test(self):
        self.service = GenericService(self.test_context, 2)
        assert True

    @ignore
    @cluster(num_nodes=2)
    def a_two_node_ignored_test(self):
        assert False

    @cluster(num_nodes=2)
    def yet_another_two_node_test(self):
        self.service = GenericService(self.test_context, 2)
        assert True

    @cluster(num_nodes=3)
    def three_node_test(self):
        self.service = GenericService(self.test_context, 3)
        assert True

    @cluster(num_nodes=3)
    def three_node_test_sleeping_30s(self):
        self.service = GenericService(self.test_context, 3)
        self.logger.warning('Sleeping for 30s')
        time.sleep(30)
        assert True

    @cluster(num_nodes=3)
    def another_three_node_test(self):
        self.service = GenericService(self.test_context, 3)
        assert True

    @cluster(num_nodes=2)
    def bad_alloc_test(self):
        # @cluster annotation specifies 2 nodes, but we ask for 3, this will fail
        self.service = GenericService(self.test_context, 3)
        time.sleep(10)
        assert True



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.cluster.cluster import ClusterNode
from ducktape.cluster.cluster_spec import ClusterSpec, NodeSpec, LINUX, WINDOWS
from ducktape.cluster.node_container import NodeContainer, NodeNotPresentError, InsufficientResourcesError, \
    InsufficientHealthyNodesError
import pytest

from ducktape.cluster.remoteaccount import RemoteAccountSSHConfig
from tests.ducktape_mock import MockAccount
from tests.runner.fake_remote_account import FakeRemoteAccount, FakeWindowsRemoteAccount


def fake_account(host, is_available=True):
    return FakeRemoteAccount(ssh_config=RemoteAccountSSHConfig(host=host), is_available=is_available)


def fake_win_account(host, is_available=True):
    return FakeWindowsRemoteAccount(ssh_config=RemoteAccountSSHConfig(host=host), is_available=is_available)


class CheckNodeContainer(object):
    def check_sizes(self):
        empty = NodeContainer()
        assert 0 == empty.size()
        assert 0 == len(empty)
        nodes = [ClusterNode(MockAccount())]
        container = NodeContainer(nodes)
        assert 1 == container.size()
        assert 1 == len(container)

    def check_add_and_remove(self):
        nodes = [ClusterNode(MockAccount()),
                 ClusterNode(MockAccount()),
                 ClusterNode(MockAccount()),
                 ClusterNode(MockAccount()),
                 ClusterNode(MockAccount())]
        container = NodeContainer([])
        assert 0 == len(container)
        container.add_node(nodes[0])
        container.add_node(nodes[1])
        container.add_node(nodes[2])
        container2 = container.clone()
        i = 0
        for node in container:
            assert nodes[i] == node
            i += 1
        assert 3 == len(container)
        container.remove_node(nodes[0])
        with pytest.raises(NodeNotPresentError):
            container.remove_node(nodes[0])
        assert 2 == len(container)
        assert 3 == len(container2)

    def check_remove_single_node_spec(self):
        """Check remove_spec() method - verify a simple happy path of removing a single node"""
        accounts = [fake_account('host1'), fake_account('host2'),
                    fake_win_account('w1'), fake_win_account('w2')]
        container = NodeContainer(accounts)
        one_linux_node_spec = ClusterSpec(nodes=[NodeSpec(LINUX)])
        one_windows_node_spec = ClusterSpec(nodes=[NodeSpec(WINDOWS)])

        def _remove_single_node(one_node_spec, os):
            assert container.can_remove_spec(one_node_spec)
            good_nodes, bad_nodes = container.remove_spec(one_node_spec)
            assert good_nodes and len(good_nodes) == 1
            assert good_nodes[0].os == os
            assert not bad_nodes

        _remove_single_node(one_windows_node_spec, WINDOWS)
        assert len(container.os_to_nodes.get(LINUX)) == 2
        assert len(container.os_to_nodes.get(WINDOWS)) == 1

        _remove_single_node(one_windows_node_spec, WINDOWS)
        assert len(container.os_to_nodes.get(LINUX)) == 2
        assert not container.os_to_nodes.get(WINDOWS)
        assert not container.can_remove_spec(one_windows_node_spec)
        with pytest.raises(InsufficientResourcesError):
            container.remove_spec(one_windows_node_spec)

        _remove_single_node(one_linux_node_spec, LINUX)
        assert len(container.os_to_nodes.get(LINUX)) == 1
        assert not container.os_to_nodes.get(WINDOWS)

        _remove_single_node(one_linux_node_spec, LINUX)
        assert not container.os_to_nodes.get(LINUX)
        assert not container.os_to_nodes.get(WINDOWS)
        assert not container.can_remove_spec(one_linux_node_spec)
        with pytest.raises(InsufficientResourcesError):
            container.remove_spec(one_linux_node_spec)

    @pytest.mark.parametrize("cluster_spec", [
        pytest.param(ClusterSpec(nodes=[NodeSpec(LINUX), NodeSpec(WINDOWS), NodeSpec(WINDOWS)]),
                     id='not enough windows nodes'),
        pytest.param(ClusterSpec(nodes=[NodeSpec(LINUX), NodeSpec(LINUX), NodeSpec(WINDOWS)]),
                     id='not enough linux nodes'),
        pytest.param(ClusterSpec(nodes=[NodeSpec(LINUX), NodeSpec(LINUX), NodeSpec(WINDOWS), NodeSpec(WINDOWS)]),
                     id='not enough nodes'),
    ])
    def check_not_enough_nodes_to_remove(self, cluster_spec):
        """
        Check what happens if there aren't enough resources in this container to match a given spec.
        Various parametrizations check the behavior for when there are enough nodes for one OS but not another,
        or for both.
        """
        accounts = [fake_account('host1'), fake_win_account('w1')]
        container = NodeContainer(accounts)
        original_container = container.clone()

        assert not container.can_remove_spec(cluster_spec)
        assert len(container.attempt_remove_spec(cluster_spec)) > 0

        with pytest.raises(InsufficientResourcesError):
            container.remove_spec(cluster_spec)

        # check that container was not modified
        assert container.os_to_nodes == original_container.os_to_nodes

    @pytest.mark.parametrize("accounts", [
        pytest.param([
            fake_account('host1'), fake_account('host2'),
            fake_win_account('w1'), fake_win_account('w2', is_available=False)
        ], id="windows not available"),
        pytest.param([
            fake_account('host1'), fake_account('host2', is_available=False),
            fake_win_account('w1'), fake_win_account('w2')
        ], id="linux not available"),
        pytest.param([
            fake_account('host1'), fake_account('host2', is_available=False),
            fake_win_account('w1'), fake_win_account('w2', is_available=False)
        ], id="neither is available")
    ])
    def check_not_enough_healthy_nodes(self, accounts):
        """
        When there's not enough healthy nodes in any of the OS-s, we obviously don't want to remove anything.
        Even when there's enough healthy nodes for one of the OS-s, but not enough in another one,
        we don't want to remove any nodes at all.
        Various sets of params check if there aren't enough healthy nodes for one OS but not the other, or both.
        """
        container = NodeContainer(accounts)
        original_container = container.clone()
        expected_bad_nodes = [acc for acc in accounts if not acc.is_available]
        spec = ClusterSpec(nodes=[NodeSpec(LINUX), NodeSpec(LINUX), NodeSpec(WINDOWS), NodeSpec(WINDOWS)])
        assert container.can_remove_spec(spec)
        with pytest.raises(InsufficientHealthyNodesError) as exc_info:
            container.remove_spec(spec)
        assert exc_info.value.bad_nodes == expected_bad_nodes
        # check that no nodes were actually allocated, but unhealthy ones were removed from the cluster
        original_container.remove_nodes(expected_bad_nodes)
        assert container.os_to_nodes == original_container.os_to_nodes

    @pytest.mark.parametrize("accounts", [
        pytest.param([
            fake_account('host1'), fake_account('host2'), fake_account('host3'),
            fake_win_account('w1', is_available=False), fake_win_account('w2'), fake_win_account('w2')
        ], id="windows not available"),
        pytest.param([
            fake_account('host1', is_available=False), fake_account('host2'), fake_account('host3'),
            fake_win_account('w1'), fake_win_account('w2'), fake_win_account('w2')
        ], id="linux not available"),
        pytest.param([
            fake_account('host1', is_available=False), fake_account('host2'), fake_account('host3'),
            fake_win_account('w1', is_available=False), fake_win_account('w2'), fake_win_account('w2')
        ], id="neither is available"),
        pytest.param([
            fake_account('host1'), fake_account('host2'), fake_account('host3'),
            fake_win_account('w1'), fake_win_account('w2'), fake_win_account('w2')
        ], id="all are available")
    ])
    def check_enough_healthy_but_some_bad_nodes_too(self, accounts):
        """
        Check that we can successfully allocate all necessary nodes - even if some nodes don't pass health checks,
        we still have enough nodes to match the provided cluster spec.

        This test assumes that if we do encounter unhealthy node,
        we encounter it before we can finish allocating healthy ones, otherwise it would just mean testing the happy
        path (which we do in one of the params).
        """
        container = NodeContainer(accounts)
        original_container = container.clone()
        expected_bad_nodes = [acc for acc in accounts if not acc.is_available]
        spec = ClusterSpec(nodes=[NodeSpec(LINUX), NodeSpec(LINUX), NodeSpec(WINDOWS), NodeSpec(WINDOWS)])

        assert container.can_remove_spec(spec)
        good_nodes, bad_nodes = container.remove_spec(spec)

        # alloc should succeed
        # check that we did catch a bad node if any
        assert bad_nodes == expected_bad_nodes
        # check that container has exactly the right number of nodes left -
        # we removed len(spec) healthy nodes, plus len(expected_bad_nodes) of unhealthy nodes.
        assert len(container) == len(original_container) - len(spec) - len(expected_bad_nodes)

        # check that we got 2 windows nodes and two linux nodes in response,
        # don't care which ones in particular
        assert len(good_nodes) == 4
        actual_linux = [node for node in good_nodes if node.os == LINUX]
        assert len(actual_linux) == 2
        actual_win = [node for node in good_nodes if node.os == WINDOWS]
        assert len(actual_win) == 2



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import collections

from ducktape.cluster.cluster_spec import ClusterSpec
from tests.ducktape_mock import FakeCluster
from ducktape.tests.scheduler import TestScheduler
from ducktape.services.service import Service

FakeContext = collections.namedtuple('FakeContext', ['test_id', 'expected_num_nodes', 'expected_cluster_spec'])


class CheckScheduler(object):
    def setup_method(self, _):
        self.cluster = FakeCluster(100)
        self.tc0 = FakeContext(0, expected_num_nodes=10, expected_cluster_spec=ClusterSpec.simple_linux(10))
        self.tc1 = FakeContext(1, expected_num_nodes=50, expected_cluster_spec=ClusterSpec.simple_linux(50))
        self.tc2 = FakeContext(2, expected_num_nodes=100, expected_cluster_spec=ClusterSpec.simple_linux(100))
        self.tc_list = [
            self.tc0,
            self.tc1,
            self.tc2,
        ]

    def check_empty(self):
        """Check expected behavior of empty scheduler."""
        scheduler = TestScheduler([], self.cluster)

        assert len(scheduler) == 0
        assert scheduler.peek() is None

    def check_non_empty_cluster_too_small(self):
        """Ensure that scheduler does not return tests if the cluster does not have enough available nodes. """

        scheduler = TestScheduler(self.tc_list, self.cluster)
        assert len(scheduler) == len(self.tc_list)
        assert scheduler.peek() is not None

        # alloc all cluster nodes so none are available
        self.cluster.alloc(Service.setup_cluster_spec(num_nodes=len(self.cluster)))
        assert self.cluster.num_available_nodes() == 0

        # peeking should not yield an object
        assert scheduler.peek() is None

    def check_simple_usage(self):
        """Check usage with fully available cluster."""

        scheduler = TestScheduler(self.tc_list, self.cluster)

        c = 2
        while len(scheduler) > 0:
            t = scheduler.peek()
            assert t.test_id == c
            assert len(scheduler) == c + 1

            scheduler.remove(t)
            assert len(scheduler) == c

            c -= 1

    def check_with_changing_cluster_availability(self):
        """Modify cluster usage in between calls to next() """

        scheduler = TestScheduler(self.tc_list, self.cluster)

        # start with 100-node cluster (configured in setup_method())
        # allocate 60 nodes; only test_id 0 (which needs 10 nodes) should be available
        nodes = self.cluster.alloc(Service.setup_cluster_spec(num_nodes=60))
        assert self.cluster.num_available_nodes() == 40
        t = scheduler.peek()
        assert t == self.tc0
        scheduler.remove(t)
        assert scheduler.peek() is None

        # return 10 nodes, so 50 are available in the cluster
        # next test from the scheduler should be test id 1 (which needs 50 nodes)
        return_nodes = nodes[: 10]
        keep_nodes = nodes[10:]
        self.cluster.free(return_nodes)
        assert self.cluster.num_available_nodes() == 50
        t = scheduler.peek()
        assert t == self.tc1
        scheduler.remove(t)
        assert scheduler.peek() is None

        # return remaining nodes, so cluster is fully available
        # next test from scheduler should be test id 2 (which needs 100 nodes)
        return_nodes = keep_nodes
        self.cluster.free(return_nodes)
        assert self.cluster.num_available_nodes() == len(self.cluster)
        t = scheduler.peek()
        assert t == self.tc2
        scheduler.remove(t)
        # scheduler should become empty now
        assert len(scheduler) == 0
        assert scheduler.peek() is None

    def check_filter_unschedulable_tests(self):
        self.cluster = FakeCluster(49)  # only test id 0 can be scheduled on this cluster
        scheduler = TestScheduler(self.tc_list, self.cluster)

        unschedulable = scheduler.filter_unschedulable_tests()
        assert set(unschedulable) == {self.tc1, self.tc2}
        # subsequent calls should return empty list and not modify scheduler
        assert not scheduler.filter_unschedulable_tests()
        assert scheduler.peek() == self.tc0



# Copyright 2014 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections


class ClusterNode(object):
    def __init__(self, account, **kwargs):
        self.account = account
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def name(self):
        return self.account.hostname

    @property
    def operating_system(self):
        return self.account.operating_system


class Cluster(object):
    """ Interface for a cluster -- a collection of nodes with login credentials.
    This interface doesn't define any mapping of roles/services to nodes. It only interacts with some underlying
    system that can describe available resources and mediates reservations of those resources.
    """

    def __init__(self):
        self.max_used_nodes = 0

    def __len__(self):
        """Size of this cluster object. I.e. number of 'nodes' in the cluster."""
        return self.available().size() + self.used().size()

    def alloc(self, cluster_spec):
        """
        Allocate some nodes.

        :param cluster_spec:                    A ClusterSpec describing the nodes to be allocated.
        :throws InsufficientResources:          If the nodes cannot be allocated.
        :return:                                Allocated nodes spec
        """
        allocated = self.do_alloc(cluster_spec)
        self.max_used_nodes = max(self.max_used_nodes, len(self.used()))
        return allocated

    def do_alloc(self, cluster_spec):
        """
        Subclasses should implement actual allocation here.

        :param cluster_spec:                    A ClusterSpec describing the nodes to be allocated.
        :throws InsufficientResources:          If the nodes cannot be allocated.
        :return:                                Allocated nodes spec
        """
        raise NotImplementedError

    def free(self, nodes):
        """Free the given node or list of nodes"""
        if isinstance(nodes, collections.Iterable):
            for s in nodes:
                self.free_single(s)
        else:
            self.free_single(nodes)

    def free_single(self, node):
        raise NotImplementedError()

    def __eq__(self, other):
        return other is not None and self.__dict__ == other.__dict__

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def num_available_nodes(self):
        return self.available().size()

    def available(self):
        """
        Return a ClusterSpec object describing the currently available nodes.
        """
        raise NotImplementedError

    def used(self):
        """
        Return a ClusterSpec object describing the currently in use nodes.
        """
        raise NotImplementedError

    def max_used(self):
        return self.max_used_nodes

    def all(self):
        """
        Return a ClusterSpec object describing all nodes.
        """
        return self.available().clone().add(self.used())



# Copyright 2014 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import boto3
import os
import base64
import winrm

from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_v1_5

from botocore.exceptions import ClientError

from ducktape.cluster.cluster_spec import WINDOWS
from ducktape.cluster.remoteaccount import RemoteAccount, RemoteCommandError


class WindowsRemoteAccount(RemoteAccount):
    """
    Windows remote accounts are currently only supported in EC2. See ``_setup_winrm()`` for how the WinRM password
    is fetched, which is currently specific to AWS.

    The Windows AMI needs to also have an SSH server running to support SSH commands, SCP, and rsync.
    """

    WINRM_USERNAME = "Administrator"

    def __init__(self, *args, **kwargs):
        super(WindowsRemoteAccount, self).__init__(*args, **kwargs)
        self.os = WINDOWS
        self._winrm_client = None

    @property
    def winrm_client(self):
        # TODO: currently this only works in AWS EC2 provisioned by Vagrant. Add support for other environments.

        # check if winrm has already been setup. If yes, return immediately.
        if self._winrm_client:
            return self._winrm_client

        # first get the instance ID of this machine from Vagrant's metadata.
        ec2_instance_id_path = os.path.join(os.getcwd(), ".vagrant", "machines", self.ssh_config.host, "aws", "id")
        instance_id_file = None
        try:
            instance_id_file = open(ec2_instance_id_path, 'r')
            ec2_instance_id = instance_id_file.read().strip()
            if not ec2_instance_id or ec2_instance_id == "":
                raise Exception
        except Exception:
            raise Exception("Could not extract EC2 instance ID from local file: %s" % ec2_instance_id_path)
        finally:
            if instance_id_file:
                instance_id_file.close()

        self._log(logging.INFO, "Found EC2 instance id: %s" % ec2_instance_id)

        # then get the encrypted password.
        client = boto3.client('ec2')
        try:
            response = client.get_password_data(InstanceId=ec2_instance_id)
        except ClientError as ce:
            if "InvalidInstanceID.NotFound" in str(ce):
                raise Exception("The instance id '%s' couldn't be found. Is the correct AWS region configured?"
                                % ec2_instance_id)
            else:
                raise ce

        self._log(logging.INFO, "Fetched encrypted winrm password and will decrypt with private key: %s"
                  % self.ssh_config.identityfile)

        # then decrypt the password using the private key.
        key_file = None
        try:
            key_file = open(self.ssh_config.identityfile, 'r')
            key = key_file.read()
            rsa_key = RSA.importKey(key)
            cipher = PKCS1_v1_5.new(rsa_key)
            winrm_password = cipher.decrypt(base64.b64decode(response["PasswordData"]), None)
            self._winrm_client = winrm.Session(self.ssh_config.hostname, auth=(WindowsRemoteAccount.WINRM_USERNAME,
                                                                               winrm_password))
        finally:
            if key_file:
                key_file.close()

        return self._winrm_client

    def fetch_externally_routable_ip(self, is_aws=None):
        # EC2 windows machines aren't given an externally routable IP. Use the hostname instead.
        return self.ssh_config.hostname

    def run_winrm_command(self, cmd, allow_fail=False):
        self._log(logging.DEBUG, "Running winrm command: %s" % cmd)
        result = self.winrm_client.run_cmd(cmd)
        if not allow_fail and result.status_code != 0:
            raise RemoteCommandError(self, cmd, result.status_code, result.std_err)
        return result.status_code



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape import __version__ as __ducktape_version__
from ducktape.errors import TimeoutError

import importlib
import time


def wait_until(condition, timeout_sec, backoff_sec=.1, err_msg="", retry_on_exc=False):
    """Block until condition evaluates as true or timeout expires, whichever comes first.

    :param condition: callable that returns True if the condition is met, False otherwise
    :param timeout_sec: number of seconds to check the condition for before failing
    :param backoff_sec: number of seconds to back off between each failure to meet the condition before checking again
    :param err_msg: a string or callable returning a string that will be included as the exception message if the
                    condition is not met
    :param retry_on_exc: if True, will retry if condition raises an exception. If condition raised exception on last
                         iteration, that exception will be raised as a cause of TimeoutError.
                         If False and condition raises an exception, that exception will be forwarded to the caller
                         immediately.
                         Defaults to False (original ducktape behavior).
                         # TODO: [1.0.0] flip this to True
    :return: silently if condition becomes true within the timeout window, otherwise raise Exception with the given
    error message.
    """
    start = time.time()
    stop = start + timeout_sec
    last_exception = None
    while time.time() < stop:
        try:
            if condition():
                return
            else:
                # reset last_exception if last iteration didn't raise any exception, but simply returned False
                last_exception = None
        except BaseException as e:
            # save last raised exception for logging it later
            last_exception = e
            if not retry_on_exc:
                raise e
        finally:
            time.sleep(backoff_sec)

    # it is safe to call Exception from None - will be just treated as a normal exception
    raise TimeoutError(err_msg() if callable(err_msg) else err_msg) from last_exception


def package_is_installed(package_name):
    """Return true iff package can be successfully imported."""
    try:
        importlib.import_module(package_name)
        return True
    except Exception:
        return False


def ducktape_version():
    """Return string representation of current ducktape version."""
    return __ducktape_version__


def load_function(func_module_path):
    """Loads and returns a function from a module path seperated by '.'s"""
    module, function = func_module_path.rsplit(".", 1)
    try:
        return getattr(importlib.import_module(module), function)
    except AttributeError:
        raise Exception("Function could not be loaded from the module path {}, "
                        "verify that it is '.' seperated".format(func_module_path))



# Copyright 2014 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.cluster.cluster_spec import LINUX
from ducktape.cluster.remoteaccount import RemoteAccount, RemoteAccountError


class LinuxRemoteAccount(RemoteAccount):

    def __init__(self, *args, **kwargs):
        super(LinuxRemoteAccount, self).__init__(*args, **kwargs)
        self._ssh_client = None
        self._sftp_client = None
        self.os = LINUX

    @property
    def local(self):
        """Returns True if this 'remote' account is probably local.
        This is an imperfect heuristic, but should work for simple local testing."""
        return self.hostname == "localhost" and self.user is None and self.ssh_config is None

    def get_network_devices(self):
        """
        Utility to get all network devices on a linux account
        """
        return [
            device
            for device in self.sftp_client.listdir('/sys/class/net')
        ]

    def get_external_accessible_network_devices(self):
        """
        gets the subset of devices accessible through an external conenction
        """
        return [
            device
            for device in self.get_network_devices()
            if device != 'lo'  # do not include local device
            and (device.startswith("en") or device.startswith('eth'))  # filter out other devices; "en" means ethernet
            # eth0 can also sometimes happen, see https://unix.stackexchange.com/q/134483
        ]

    # deprecated, please use the self.externally_routable_ip that is set in your cluster,
    # not explicitly deprecating it as it's used by vagrant cluster
    def fetch_externally_routable_ip(self, is_aws=None):
        if is_aws is not None:
            self.logger.warning("fetch_externally_routable_ip: is_aws is a deprecated flag, and does nothing")

        devices = self.get_external_accessible_network_devices()

        self.logger.debug("found devices: {}".format(devices))

        if not devices:
            raise RemoteAccountError(self, "Couldn't find any network devices")

        fmt_cmd = (
            "/sbin/ifconfig {device} | "
            "grep 'inet ' | "
            "tail -n 1 | "
            r"egrep -o '[0-9\.]+' | "
            "head -n 1 2>&1"
        )

        ips = [
            "".join(
                self.ssh_capture(fmt_cmd.format(device=device))
            ).strip()
            for device in devices
        ]
        self.logger.debug("found ips: {}".format(ips))
        self.logger.debug("returning the first ip found")
        return next(iter(ips))



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import json
from pathlib import Path

import yaml
import os
import shutil
import xml.etree.ElementTree as ET

import pkg_resources

from ducktape.utils.terminal_size import get_terminal_size
from ducktape.utils.util import ducktape_version
from ducktape.tests.status import PASS, FAIL, IGNORE, FLAKY
from ducktape.json_serializable import DucktapeJSONEncoder


DEFAULT_SEPARATOR_WIDTH = 100


def format_time(t):
    """Return human-readable interval of time.
    Assumes t is in units of seconds.
    """
    minutes = int(t / 60)
    seconds = t % 60

    r = ""
    if minutes > 0:
        r += "%d minute%s " % (minutes, "" if minutes == 1 else "s")
    r += "%.3f seconds" % seconds
    return r


class SingleResultReporter(object):
    """Helper class for creating a view of results from a single test."""

    def __init__(self, result):
        self.result = result
        self.width = get_terminal_size()[0]

    def result_string(self):
        """Stringify single result"""
        result_lines = [
            "test_id:    %s" % self.result.test_id,
            "status:     %s" % str(self.result.test_status).upper(),
            "run time:   %s" % format_time(self.result.run_time_seconds),
        ]

        if self.result.test_status == FAIL:
            # Add summary if the test failed
            result_lines.append("\n")
            result_lines.append("    " + self.result.summary)

        if self.result.data is not None:
            result_lines.append(json.dumps(self.result.data))

        return "\n".join(result_lines)

    def report_string(self):
        """Get the whole report string."""
        return "\n".join(["=" * self.width,
                          self.result_string()])


class SingleResultFileReporter(SingleResultReporter):
    def report(self):
        self.width = DEFAULT_SEPARATOR_WIDTH
        report_file = os.path.join(self.result.results_dir, "report.txt")
        with open(report_file, "w") as fp:
            fp.write(self.report_string())

        # write collected data
        if self.result.data is not None:
            data_file = os.path.join(self.result.results_dir, "data.json")
            with open(data_file, "w") as fp:
                fp.write(json.dumps(self.result.data))


class SummaryReporter(object):
    def __init__(self, results):
        self.results = results
        self.width = get_terminal_size()[0]

    def report(self):
        raise NotImplementedError("method report must be implemented by subclasses of SummaryReporter")


class SimpleSummaryReporter(SummaryReporter):
    def header_string(self):
        """Header lines of the report"""
        header_lines = [
            "=" * self.width,
            "SESSION REPORT (ALL TESTS)",
            "ducktape version: %s" % ducktape_version(),
            "session_id:       %s" % self.results.session_context.session_id,
            "run time:         %s" % format_time(self.results.run_time_seconds),
            "tests run:        %d" % len(self.results),
            "passed:           %d" % self.results.num_passed,
            "flaky:            %d" % self.results.num_flaky,
            "failed:           %d" % self.results.num_failed,
            "ignored:          %d" % self.results.num_ignored,
            "=" * self.width
        ]

        return "\n".join(header_lines)

    def report_string(self):
        """Get the whole report string."""
        report_lines = [
            self.header_string()]

        report_lines.extend(
            [SingleResultReporter(result).result_string() + "\n" + "-" * self.width for result in self.results])

        return "\n".join(report_lines)


class SimpleFileSummaryReporter(SimpleSummaryReporter):
    def report(self):
        self.width = DEFAULT_SEPARATOR_WIDTH
        report_file = os.path.join(self.results.session_context.results_dir, "report.txt")
        with open(report_file, "w") as fp:
            fp.write(self.report_string())


class SimpleStdoutSummaryReporter(SimpleSummaryReporter):
    def report(self):
        print(self.report_string())


class JSONReporter(object):
    def __init__(self, results):
        self.results = results

    def report(self):
        report_file = os.path.abspath(os.path.join(self.results.session_context.results_dir, "report.json"))
        with open(report_file, "w") as f:
            f.write(json.dumps(self.results, cls=DucktapeJSONEncoder, sort_keys=True, indent=2, separators=(',', ': ')))


class JUnitReporter(object):
    def __init__(self, results):
        self.results = results

    def report(self):
        report_file = os.path.abspath(os.path.join(self.results.session_context.results_dir, "report.xml"))
        testsuites = {}

        # First bucket by module_name and argregate counts
        for result in self.results:
            module_name = result.module_name
            testsuites.setdefault(module_name, {})
            # Set default values
            testsuite = testsuites[module_name]
            testsuite.setdefault('tests', 0)
            testsuite.setdefault('skipped', 0)
            testsuite.setdefault('failures', 0)
            testsuite.setdefault('errors', 0)
            testsuite.setdefault('testcases', []).append(result)

            # Always increment total number of tests
            testsuite['tests'] += 1
            if result.test_status == FAIL:
                testsuite['failures'] += 1
            elif result.test_status == IGNORE:
                testsuite['skipped'] += 1

        total = self.results.num_failed + self.results.num_ignored + self.results.num_passed + self.results.num_flaky
        # Now start building XML document
        root = ET.Element('testsuites', attrib=dict(
            name="ducktape", time=str(self.results.run_time_seconds),
            tests=str(total), disabled="0", errors="0",
            failures=str(self.results.num_failed)
        ))
        for module_name, testsuite in testsuites.items():
            xml_testsuite = ET.SubElement(root, 'testsuite', attrib=dict(
                name=module_name, tests=str(testsuite['tests']), disabled="0",
                errors="0", failures=str(testsuite['failures']), skipped=str(testsuite['skipped'])
            ))
            for test in testsuite['testcases']:
                # Since we're already aware of module_name and cls_name, strip that prefix off
                full_name = "{module_name}.{cls_name}.".format(module_name=module_name, cls_name=test.cls_name)
                if test.test_id.startswith(full_name):
                    name = test.test_id[len(full_name):]
                else:
                    name = test.test_id
                xml_testcase = ET.SubElement(xml_testsuite, 'testcase', attrib=dict(
                    name=name, classname=test.cls_name, time=str(test.run_time_seconds),
                    status=str(test.test_status), assertions=""
                ))
                if test.test_status == FAIL:
                    xml_failure = ET.SubElement(xml_testcase, 'failure', attrib=dict(
                        message=test.summary.splitlines()[0]
                    ))
                    xml_failure.text = test.summary
                elif test.test_status == IGNORE:
                    ET.SubElement(xml_testcase, 'skipped')

        with open(report_file, "w") as f:
            content = ET.tostring(root)
            if isinstance(content, bytes):
                content = content.decode("utf-8")
            f.write(content)


class HTMLSummaryReporter(SummaryReporter):

    def format_test_name(self, result):
        lines = ["Module: " + result.module_name,
                 "Class:  " + result.cls_name,
                 "Method: " + result.function_name]

        if result.injected_args is not None:
            lines.append("Arguments:")
            lines.append(
                json.dumps(result.injected_args, sort_keys=True, indent=2, separators=(',', ': ')))

        return "\n".join(lines)

    def format_result(self, result):
        test_result = str(result.test_status).lower()

        result_json = {
            "test_name": self.format_test_name(result),
            "test_result": test_result,
            "description": result.description,
            "run_time": format_time(result.run_time_seconds),
            "data": "" if result.data is None else json.dumps(result.data, sort_keys=True,
                                                              indent=2, separators=(',', ': ')),
            "summary": result.summary,
            "test_log": self.test_results_dir(result)
        }
        return result_json

    def test_results_dir(self, result):
        """Return *relative path* to test results directory.

        Path is relative to the base results_dir. Relative path behaves better if the results directory is copied,
        moved etc.
        """
        base_dir = os.path.abspath(result.session_context.results_dir)
        base_dir = os.path.join(base_dir, "")  # Ensure trailing directory indicator

        test_results_dir = os.path.abspath(result.results_dir)
        return test_results_dir[len(base_dir):]  # truncate the "absolute" portion

    def format_report(self):
        template = pkg_resources.resource_string(__name__, '../templates/report/report.html').decode('utf-8')

        num_tests = len(self.results)
        num_passes = 0
        failed_result_string = []
        passed_result_string = []
        ignored_result_string = []
        flaky_result_string = []

        for result in self.results:
            json_string = json.dumps(self.format_result(result))
            if result.test_status == PASS:
                num_passes += 1
                passed_result_string.append(json_string)
                passed_result_string.append(",")
            elif result.test_status == FAIL:
                failed_result_string.append(json_string)
                failed_result_string.append(",")
            elif result.test_status == IGNORE:
                ignored_result_string.append(json_string)
                ignored_result_string.append(",")
            elif result.test_status == FLAKY:
                flaky_result_string.append(json_string)
                flaky_result_string.append(",")
            else:
                raise Exception("Unknown test status in report: {}".format(result.test_status.to_json()))

        args = {
            'ducktape_version': ducktape_version(),
            'num_tests': num_tests,
            'num_passes': self.results.num_passed,
            'num_flaky': self.results.num_flaky,
            'num_failures': self.results.num_failed,
            'num_ignored': self.results.num_ignored,
            'run_time': format_time(self.results.run_time_seconds),
            'session': self.results.session_context.session_id,
            'passed_tests': "".join(passed_result_string),
            'flaky_tests': "".join(flaky_result_string),
            'failed_tests': "".join(failed_result_string),
            'ignored_tests': "".join(ignored_result_string),
            'test_status_names': ",".join(["\'%s\'" % str(status) for status in [PASS, FAIL, IGNORE, FLAKY]])
        }

        html = template % args
        report_html = os.path.join(self.results.session_context.results_dir, "report.html")
        with open(report_html, "w") as fp:
            fp.write(html)
            fp.close()

        report_css = os.path.join(self.results.session_context.results_dir, "report.css")
        report_css_origin = pkg_resources.resource_filename(__name__, '../templates/report/report.css')
        shutil.copy2(report_css_origin, report_css)

    def report(self):
        self.format_report()


class FailedTestSymbolReporter(SummaryReporter):

    def __init__(self, results):
        super().__init__(results)
        self.working_dir = Path().absolute()
        self.separator = "=" * self.width

    def to_symbol(self, result):
        p = Path(result.file_name).relative_to(self.working_dir)
        line = f'{p}::{result.cls_name}.{result.function_name}'
        if result.injected_args:
            injected_args_str = json.dumps(result.injected_args, separators=(',', ':'))
            line += f'@{injected_args_str}'
        return line

    def dump_test_suite(self, lines):
        print(self.separator)
        print('FAILED TEST SUITE')
        suite = {self.results.session_context.session_id: lines}
        file_path = Path(self.results.session_context.results_dir) / "rerun-failed.yml"
        with file_path.open('w') as fp:
            print(f'Test suite to rerun failed tests: {file_path}')
            yaml.dump(suite, stream=fp, indent=4)

    def print_test_symbols_string(self, lines):
        print(self.separator)
        print('FAILED TEST SYMBOLS')
        print('Pass the test symbols below to your ducktape run')
        # quote the symbol because json parameters will be processed by shell otherwise, making it not copy-pasteable
        print(' '.join([f"'{line}'" for line in lines]))

    def report(self):
        symbols = [self.to_symbol(result) for result in self.results if result.test_status == FAIL]
        if not symbols:
            return

        self.dump_test_suite(symbols)
        self.print_test_symbols_string(symbols)
